# SPDX-License-Identifier: GPL-2.0-only
# This file is part of Scapy
# See https://scapy.net/ for more information
# Copyright (C) Nils Weiss <nils@we155.de>

# scapy.contrib.status = skip

"""
Package of contrib automotive modules that have to be loaded explicitly.
"""

import logging

log_automotive = logging.getLogger("scapy.contrib.automotive")

log_automotive.setLevel(logging.INFO)



# SPDX-License-Identifier: GPL-2.0-only
# This file is part of Scapy
# See https://scapy.net/ for more information
# Copyright (C) Guillaume Valadon <guillaume@valadon.net>

"""
Scapy BSD native support
"""



# SPDX-License-Identifier: GPL-2.0-or-later
# This file is part of Scapy
# See https://scapy.net/ for more information
# Copyright (C) 2020 Michael Farrell <micolous+git@gmail.com>

# scapy.contrib.description = Apple/Google Exposure Notification System (ENS)
# scapy.contrib.status = loads

"""
Apple/Google Exposure Notification System (ENS), formerly known as
Privacy-Preserving Contact Tracing Project.

This module parses the Bluetooth Low Energy beacon payloads used by the system.
This does **not** yet implement any cryptographic functionality.

More info:

* `Apple: Privacy-Preserving Contact Tracing`__
* `Google: Exposure Notifications`__
* `Wikipedia: Exposure Notification`__

__ https://www.apple.com/covid19/contacttracing/
__ https://www.google.com/covid19/exposurenotifications/
__ https://en.wikipedia.org/wiki/Exposure_Notification

Bluetooth protocol specifications:

* `v1.1`_ (April 2020)
* `v1.2`_ (April 2020)

.. _v1.1: https://blog.google/documents/58/Contact_Tracing_-_Bluetooth_Specification_v1.1_RYGZbKW.pdf
.. _v1.2: https://covid19-static.cdn-apple.com/applications/covid19/current/static/contact-tracing/pdf/ExposureNotification-BluetoothSpecificationv1.2.pdf
"""  # noqa: E501

from scapy.fields import StrFixedLenField
from scapy.layers.bluetooth import EIR_Hdr, EIR_ServiceData16BitUUID, \
    EIR_CompleteList16BitServiceUUIDs, LowEnergyBeaconHelper
from scapy.packet import bind_layers, Packet


EXPOSURE_NOTIFICATION_UUID = 0xFD6F


class Exposure_Notification_Frame(Packet, LowEnergyBeaconHelper):
    """Apple/Google BLE Exposure Notification broadcast frame."""
    name = "Exposure Notification broadcast"

    fields_desc = [
        # Rolling Proximity Identifier
        StrFixedLenField("identifier", None, 16),
        # Associated Encrypted Metadata (added in v1.2)
        StrFixedLenField("metadata", None, 4),
    ]

    def build_eir(self):
        """Builds a list of EIR messages to wrap this frame."""

        return LowEnergyBeaconHelper.base_eir + [
            EIR_Hdr() / EIR_CompleteList16BitServiceUUIDs(svc_uuids=[
                EXPOSURE_NOTIFICATION_UUID]),
            EIR_Hdr() / EIR_ServiceData16BitUUID() / self
        ]


bind_layers(EIR_ServiceData16BitUUID, Exposure_Notification_Frame,
            svc_uuid=EXPOSURE_NOTIFICATION_UUID)



# SPDX-License-Identifier: GPL-2.0-only
# This file is part of Scapy
# See https://scapy.net/ for more information
# Copyright (C) Nils Weiss <nils@we155.de>

# scapy.contrib.description = ISO-TP (ISO 15765-2) Packet Definitions
# scapy.contrib.status = library

import logging
import struct
# Typing imports
from typing import (
    Optional,
    List,
    Tuple,
    Any,
    Type,
    cast,
)

from scapy.compat import chb, orb
from scapy.config import conf
from scapy.error import Scapy_Exception
from scapy.fields import BitField, FlagsField, StrLenField, \
    ThreeBytesField, XBitField, ConditionalField, \
    BitEnumField, ByteField, XByteField, BitFieldLenField, StrField, \
    FieldLenField, IntField, ShortField
from scapy.layers.can import CAN, CAN_FD_MAX_DLEN as CAN_FD_MAX_DLEN, CANFD
from scapy.packet import Packet

log_isotp = logging.getLogger("scapy.contrib.isotp")

CAN_MAX_IDENTIFIER = (1 << 29) - 1  # Maximum 29-bit identifier
CAN_MTU = 16
CAN_MAX_DLEN = 8
ISOTP_MAX_DLEN_2015 = (1 << 32) - 1  # Maximum for 32-bit FF_DL
ISOTP_MAX_DLEN = (1 << 12) - 1  # Maximum for 12-bit FF_DL
ISOTP_TYPES = {0: 'single',
               1: 'first',
               2: 'consecutive',
               3: 'flow_control'}

N_PCI_SF = 0x00  # /* single frame */
N_PCI_FF = 0x10  # /* first frame */
N_PCI_CF = 0x20  # /* consecutive frame */
N_PCI_FC = 0x30  # /* flow control */


class ISOTP(Packet):
    """Packet class for ISOTP messages. This class contains additional
    slots for source address (tx_id), destination address (rx_id),
    extended source address (ext_address) and
    extended destination address (rx_ext_address) information. This information
    gets filled from ISOTPSockets or the ISOTPMessageBuilder, if it
    is available. Address information is not used for Packet comparison.

    :param args: Arguments for Packet init, for example bytes string
    :param kwargs: Keyword arguments for Packet init.
    """
    name = 'ISOTP'
    fields_desc = [
        StrField('data', b"")
    ]
    __slots__ = Packet.__slots__ + ["tx_id", "rx_id", "ext_address", "rx_ext_address"]  # noqa: E501

    def __init__(self, *args, **kwargs):
        # type: (Any, Any) -> None
        self.tx_id = kwargs.pop("tx_id", None)  # type: Optional[int]
        self.rx_id = kwargs.pop("rx_id", None)  # type: Optional[int]
        self.ext_address = kwargs.pop("ext_address", None)  # type: Optional[int]  # noqa: E501
        self.rx_ext_address = kwargs.pop("rx_ext_address", None)  # type: Optional[int]  # noqa: E501
        Packet.__init__(self, *args, **kwargs)
        self.validate_fields()

    def validate_fields(self):
        # type: () -> None
        """Helper function to validate information in tx_id, rx_id,
        ext_address and rx_ext_address slots
        """
        if self.tx_id is not None:
            if not 0 <= self.tx_id <= CAN_MAX_IDENTIFIER:
                raise Scapy_Exception("tx_id is not a valid CAN identifier")
        if self.rx_id is not None:
            if not 0 <= self.rx_id <= CAN_MAX_IDENTIFIER:
                raise Scapy_Exception("rx_id is not a valid CAN identifier")
        if self.ext_address is not None:
            if not 0 <= self.ext_address <= 0xff:
                raise Scapy_Exception("ext_address is not a byte")
        if self.rx_ext_address is not None:
            if not 0 <= self.rx_ext_address <= 0xff:
                raise Scapy_Exception("rx_ext_address is not a byte")

    def fragment(self, *args, **kargs):
        # type: (*Any, **Any) -> List[Packet]
        """Helper function to fragment an ISOTP message into multiple
        CAN frames.

        :param fd: type: Optional[bool]: will fragment the can frames
            with size CAN_FD_MAX_DLEN

        :return: A list of CAN frames
        """

        fd = kargs.pop("fd", False)
        pkt_cls = CANFD if fd else CAN

        def _get_data_len():
            # type: () -> int
            return CAN_MAX_DLEN if not fd else CAN_FD_MAX_DLEN

        data_bytes_in_frame = _get_data_len() - 1
        if self.rx_ext_address is not None:
            data_bytes_in_frame = data_bytes_in_frame - 1

        if len(self.data) > ISOTP_MAX_DLEN_2015:
            raise Scapy_Exception("Too much data in ISOTP message")

        if len(self.data) <= data_bytes_in_frame:
            # We can do this in a single frame
            frame_data = struct.pack('B', len(self.data)) + self.data
            if self.rx_ext_address:
                frame_data = struct.pack('B', self.rx_ext_address) + frame_data

            if self.rx_id is None or self.rx_id <= 0x7ff:
                pkt = pkt_cls(identifier=self.rx_id, data=frame_data)
            else:
                pkt = pkt_cls(identifier=self.rx_id, flags="extended",
                              data=frame_data)
            return [pkt]

        # Construct the first frame
        if len(self.data) <= ISOTP_MAX_DLEN:
            frame_header = struct.pack(">H", len(self.data) + 0x1000)
        else:
            frame_header = struct.pack(">HI", 0x1000, len(self.data))
        if self.rx_ext_address:
            frame_header = struct.pack('B', self.rx_ext_address) + frame_header
        idx = _get_data_len() - len(frame_header)
        frame_data = self.data[0:idx]
        if self.rx_id is None or self.rx_id <= 0x7ff:
            frame = pkt_cls(identifier=self.rx_id, data=frame_header + frame_data)
        else:
            frame = pkt_cls(identifier=self.rx_id, flags="extended",
                            data=frame_header + frame_data)

        # Construct consecutive frames
        n = 1
        pkts = [frame]
        while idx < len(self.data):
            frame_data = self.data[idx:idx + data_bytes_in_frame]
            frame_header = struct.pack("b", (n % 16) + N_PCI_CF)

            n += 1
            idx += len(frame_data)

            if self.rx_ext_address:
                frame_header = struct.pack('B', self.rx_ext_address) + frame_header  # noqa: E501
            if self.rx_id is None or self.rx_id <= 0x7ff:
                pkt = pkt_cls(identifier=self.rx_id, data=frame_header + frame_data)  # noqa: E501
            else:
                pkt = pkt_cls(identifier=self.rx_id, flags="extended",
                              data=frame_header + frame_data)
            pkts.append(pkt)
        return cast(List[Packet], pkts)

    @staticmethod
    def defragment(can_frames, use_extended_addressing=None):
        # type: (List[Packet], Optional[bool]) -> Optional[ISOTP]
        """Helper function to defragment a list of CAN frames to one ISOTP
        message

        :param can_frames: A list of CAN frames
        :param use_extended_addressing: Specify if extended ISO-TP addressing
                                        is used in the packets for
                                        defragmentation.
        :return: An ISOTP message containing the data of the CAN frames or None
        """
        from scapy.contrib.isotp.isotp_utils import ISOTPMessageBuilder

        if len(can_frames) == 0:
            raise Scapy_Exception("ISOTP.defragment called with 0 frames")

        dst = can_frames[0].identifier
        if any(frame.identifier != dst for frame in can_frames):
            log_isotp.warning("Not all CAN frames have the same identifier")

        parser = ISOTPMessageBuilder(use_extended_addressing)
        parser.feed(can_frames)

        results = []
        for p in parser:
            if (use_extended_addressing is True and
                p.rx_ext_address is not None) \
                    or (use_extended_addressing is False and
                        p.rx_ext_address is None) \
                    or (use_extended_addressing is None):
                results.append(p)

        if not results:
            return None

        if len(results) > 1:
            log_isotp.warning(
                "More than one ISOTP frame could be defragmented from the "
                "provided CAN frames, only returning the first one.")

        return results[0]


class ISOTPHeader(CAN):
    name = 'ISOTPHeader'
    fields_desc = [
        FlagsField('flags', 0, 3, ['error',
                                   'remote_transmission_request',
                                   'extended']),
        XBitField('identifier', 0, 29),
        ByteField('length', None),
        ThreeBytesField('reserved', 0)
    ]

    def extract_padding(self, p):
        # type: (bytes) -> Tuple[bytes, Optional[bytes]]
        return p, None

    def post_build(self, pkt, pay):
        # type: (bytes, bytes) -> bytes
        """
        This will set the ByteField 'length' to the correct value.
        """
        if self.length is None:
            pkt = pkt[:4] + chb(len(pay)) + pkt[5:]

        if conf.contribs['CAN']['swap-bytes']:
            data = CAN.inv_endianness(pkt)  # type: bytes
            return data + pay
        return pkt + pay

    def guess_payload_class(self, payload):
        # type: (bytes) -> Type[Packet]
        """ISO-TP encodes the frame type in the first nibble of a frame. This
        is used to determine the payload_class

        :param payload: payload bytes string
        :return: Type of payload class
        """
        if len(payload) < 1:
            return self.default_payload_class(payload)

        t = (orb(payload[0]) & 0xf0) >> 4
        if t == 0:
            length = (orb(payload[0]) & 0x0f)
            if length == 0:
                return ISOTP_SF_FD
            else:
                return ISOTP_SF
        elif t == 1:
            if len(payload) < 2:
                return self.default_payload_class(payload)
            length = ((orb(payload[0]) & 0x0f) << 12) + orb(payload[1])
            if length == 0:
                return ISOTP_FF_FD
            else:
                return ISOTP_FF
        elif t == 2:
            return ISOTP_CF
        else:
            return ISOTP_FC


class ISOTPHeader_FD(ISOTPHeader):
    name = 'ISOTPHeaderFD'
    fields_desc = [
        FlagsField('flags', 0, 3, ['error',
                                   'remote_transmission_request',
                                   'extended']),
        XBitField('identifier', 0, 29),
        ByteField('length', None),
        FlagsField('fd_flags', 4, 8, ['bit_rate_switch',
                                      'error_state_indicator',
                                      'fd_frame']),
        ShortField('reserved', 0),
    ]

    def post_build(self, pkt, pay):
        # type: (bytes, bytes) -> bytes

        data = super().post_build(pkt, pay)

        length = data[4]

        if 8 < length <= 24:
            wire_length = length + (-length) % 4
        elif 24 < length <= 64:
            wire_length = length + (-length) % 8
        elif length > 64:
            raise NotImplementedError
        else:
            wire_length = length

        pad = b"\x00" * (wire_length - length)

        return data[0:4] + chb(wire_length) + data[5:] + pad


class ISOTPHeaderEA(ISOTPHeader):
    name = 'ISOTPHeaderExtendedAddress'
    fields_desc = [
        FlagsField('flags', 0, 3, ['error',
                                   'remote_transmission_request',
                                   'extended']),
        XBitField('identifier', 0, 29),
        ByteField('length', None),
        ThreeBytesField('reserved', 0),
        XByteField('extended_address', 0)
    ]

    def post_build(self, pkt, pay):
        # type: (bytes, bytes) -> bytes
        """
        This will set the ByteField 'length' to the correct value.
        'chb(len(pay) + 1)' is required, because the field 'extended_address'
        is counted as payload on the CAN layer
        """
        if self.length is None:
            pkt = pkt[:4] + chb(len(pay) + 1) + pkt[5:]

        if conf.contribs['CAN']['swap-bytes']:
            data = CAN.inv_endianness(pkt)  # type: bytes
            return data + pay
        return pkt + pay


class ISOTPHeaderEA_FD(ISOTPHeaderEA):
    name = 'ISOTPHeaderExtendedAddressFD'
    fields_desc = [
        FlagsField('flags', 0, 3, ['error',
                                   'remote_transmission_request',
                                   'extended']),
        XBitField('identifier', 0, 29),
        ByteField('length', None),
        FlagsField('fd_flags', 4, 8, ['bit_rate_switch',
                                      'error_state_indicator',
                                      'fd_frame']),
        ShortField('reserved', 0),
        XByteField('extended_address', 0)
    ]

    def post_build(self, pkt, pay):
        # type: (bytes, bytes) -> bytes

        data = super().post_build(pkt, pay)

        length = data[4]

        if 8 < length <= 24:
            wire_length = length + (-length) % 4
        elif 24 < length <= 64:
            wire_length = length + (-length) % 8
        elif length > 64:
            raise NotImplementedError
        else:
            wire_length = length

        pad = b"\x00" * (wire_length - length)

        return data[0:4] + chb(wire_length) + data[5:] + pad


ISOTP_TYPE = {0: 'single',
              1: 'first',
              2: 'consecutive',
              3: 'flow_control'}


class ISOTP_SF(Packet):
    name = 'ISOTPSingleFrame'
    fields_desc = [
        BitEnumField('type', 0, 4, ISOTP_TYPE),
        BitFieldLenField('message_size', None, 4, length_of='data'),
        StrLenField('data', b'', length_from=lambda pkt: pkt.message_size)
    ]


class ISOTP_SF_FD(Packet):
    name = 'ISOTPSingleFrameFD'
    fields_desc = [
        BitEnumField('type', 0, 4, ISOTP_TYPE),
        BitField('zero_field', 0, 4),
        FieldLenField('message_size', None, length_of='data', fmt="B"),
        StrLenField('data', b'', length_from=lambda pkt: pkt.message_size)
    ]


class ISOTP_FF(Packet):
    name = 'ISOTPFirstFrame'
    fields_desc = [
        BitEnumField('type', 1, 4, ISOTP_TYPE),
        BitField('message_size', 0, 12),
        ConditionalField(IntField('extended_message_size', 0),
                         lambda pkt: pkt.message_size == 0),
        StrField('data', b'', fmt="B")
    ]


class ISOTP_FF_FD(Packet):
    name = 'ISOTPFirstFrame'
    fields_desc = [
        BitEnumField('type', 1, 4, ISOTP_TYPE),
        BitField('zero_field', 0, 12),
        IntField('message_size', 0),
        StrField('data', b'', fmt="B")
    ]


class ISOTP_CF(Packet):
    name = 'ISOTPConsecutiveFrame'
    fields_desc = [
        BitEnumField('type', 2, 4, ISOTP_TYPE),
        BitField('index', 0, 4),
        StrField('data', b'', fmt="B")
    ]


class ISOTP_FC(Packet):
    name = 'ISOTPFlowControlFrame'
    fields_desc = [
        BitEnumField('type', 3, 4, ISOTP_TYPE),
        BitEnumField('fc_flag', 0, 4, {0: 'continue',
                                       1: 'wait',
                                       2: 'abort'}),
        ByteField('block_size', 0),
        ByteField('separation_time', 0),
    ]



# SPDX-License-Identifier: GPL-2.0-only
# This file is part of Scapy
# See https://scapy.net/ for more information
# Copyright (C) Philippe Biondi <phil@secdev.org>

"""
ISAKMP (Internet Security Association and Key Management Protocol).
"""

# Mostly based on https://tools.ietf.org/html/rfc2408

import struct
from scapy.config import conf
from scapy.packet import Packet, bind_bottom_up, bind_top_down, bind_layers
from scapy.compat import chb
from scapy.fields import (
    ByteEnumField,
    ByteField,
    FieldLenField,
    FieldListField,
    FlagsField,
    IPField,
    IntEnumField,
    IntField,
    MultipleTypeField,
    PacketLenField,
    ShortEnumField,
    ShortField,
    StrLenEnumField,
    StrLenField,
    XByteField,
    XStrFixedLenField,
    XStrLenField,
)
from scapy.layers.inet import IP, UDP
from scapy.layers.ipsec import NON_ESP
from scapy.sendrecv import sr
from scapy.volatile import RandString
from scapy.error import warning
from functools import reduce

# TODO: some ISAKMP payloads are not implemented,
# and inherit a default ISAKMP_payload


# see https://www.iana.org/assignments/ipsec-registry/ipsec-registry.xhtml#ipsec-registry-2 for details  # noqa: E501
ISAKMPAttributeTypes = {
    "Encryption": (1, {"DES-CBC": 1,
                       "IDEA-CBC": 2,
                       "Blowfish-CBC": 3,
                       "RC5-R16-B64-CBC": 4,
                       "3DES-CBC": 5,
                       "CAST-CBC": 6,
                       "AES-CBC": 7,
                       "CAMELLIA-CBC": 8, }, 0),
    "Hash": (2, {"MD5": 1,
                 "SHA": 2,
                 "Tiger": 3,
                 "SHA2-256": 4,
                 "SHA2-384": 5,
                 "SHA2-512": 6, }, 0),
    "Authentication": (3, {"PSK": 1,
                           "DSS": 2,
                           "RSA Sig": 3,
                           "RSA Encryption": 4,
                           "RSA Encryption Revised": 5,
                           "ElGamal Encryption": 6,
                           "ElGamal Encryption Revised": 7,
                           "ECDSA Sig": 8,
                           "HybridInitRSA": 64221,
                           "HybridRespRSA": 64222,
                           "HybridInitDSS": 64223,
                           "HybridRespDSS": 64224,
                           "XAUTHInitPreShared": 65001,
                           "XAUTHRespPreShared": 65002,
                           "XAUTHInitDSS": 65003,
                           "XAUTHRespDSS": 65004,
                           "XAUTHInitRSA": 65005,
                           "XAUTHRespRSA": 65006,
                           "XAUTHInitRSAEncryption": 65007,
                           "XAUTHRespRSAEncryption": 65008,
                           "XAUTHInitRSARevisedEncryption": 65009,  # noqa: E501
                           "XAUTHRespRSARevisedEncryptio": 65010, }, 0),  # noqa: E501
    "GroupDesc": (4, {"768MODPgr": 1,
                      "1024MODPgr": 2,
                      "EC2Ngr155": 3,
                      "EC2Ngr185": 4,
                      "1536MODPgr": 5,
                      "2048MODPgr": 14,
                      "3072MODPgr": 15,
                      "4096MODPgr": 16,
                      "6144MODPgr": 17,
                      "8192MODPgr": 18, }, 0),
    "GroupType": (5, {"MODP": 1,
                      "ECP": 2,
                      "EC2N": 3}, 0),
    "GroupPrime": (6, {}, 1),
    "GroupGenerator1": (7, {}, 1),
    "GroupGenerator2": (8, {}, 1),
    "GroupCurveA": (9, {}, 1),
    "GroupCurveB": (10, {}, 1),
    "LifeType": (11, {"Seconds": 1,
                      "Kilobytes": 2}, 0),
    "LifeDuration": (12, {}, 1),
    "PRF": (13, {}, 0),
    "KeyLength": (14, {}, 0),
    "FieldSize": (15, {}, 0),
    "GroupOrder": (16, {}, 1),
}

# see https://www.iana.org/assignments/isakmp-registry/isakmp-registry.xhtml#isakmp-registry-13 for details  # noqa: E501
IPSECAttributeTypes = {
    "LifeType": (1, {"Reserved": 0,
                     "seconds": 1,
                     "kilobytes": 2}, 0),
    "LifeDuration": (2, {}, 1),
    "GroupDesc": (3, ISAKMPAttributeTypes["GroupDesc"][1], 0),
    "EncapsulationMode": (4, {"Reserved": 0,
                              "Tunnel": 1,
                              "Transport": 2,
                              "UDP-Encapsulated-Tunnel": 3,
                              "UDP-Encapsulated-Transport": 4}, 0),
    "AuthenticationAlgorithm": (5, {"HMAC-MD5": 1,
                                    "HMAC-SHA": 2,
                                    "DES-MAC": 3,
                                    "KPDK": 4,
                                    "HMAC-SHA2-256": 5,
                                    "HMAC-SHA2-384": 6,
                                    "HMAC-SHA2-512": 7,
                                    "HMAC-RIPEMD": 8,
                                    "AES-XCBC-MAC": 9,
                                    "SIG-RSA": 10,
                                    "AES-128-GMAC": 11,
                                    "AES-192-GMAC": 12,
                                    "AES-256-GMAC": 13}, 0),
    "KeyLength": (6, {}, 0),
    "KeyRounds": (7, {}, 0),
    "CompressDictionarySize": (8, {}, 0),
    "CompressPrivateAlgorithm": (9, {}, 1),
}

_rev = lambda x: {
    v[0]: (k, {vv: kk for kk, vv in v[1].items()}, v[2])
    for k, v in x.items()
}
ISAKMPTransformNum = _rev(ISAKMPAttributeTypes)
IPSECTransformNum = _rev(IPSECAttributeTypes)

# See IPSEC Security Protocol Identifiers entry in
# https://www.iana.org/assignments/isakmp-registry/isakmp-registry.xhtml#isakmp-registry-3
PROTO_ISAKMP = 1
PROTO_IPSEC_AH = 2
PROTO_IPSEC_ESP = 3
PROTO_IPCOMP = 4
PROTO_GIGABEAM_RADIO = 5


class ISAKMPTransformSetField(StrLenField):
    islist = 1

    @staticmethod
    def type2num(type_val_tuple, proto=0):
        typ, val = type_val_tuple
        if proto == PROTO_ISAKMP:
            type_val, enc_dict, tlv = ISAKMPAttributeTypes.get(typ, (typ, {}, 0))
        elif proto == PROTO_IPSEC_ESP:
            type_val, enc_dict, tlv = IPSECAttributeTypes.get(typ, (typ, {}, 0))
        else:
            type_val, enc_dict, tlv = (typ, {}, 0)
        val = enc_dict.get(val, val)
        if isinstance(val, str):
            raise ValueError("Unknown attribute '%s'" % val)
        s = b""
        if (val & ~0xffff):
            if not tlv:
                warning("%r should not be TLV but is too big => using TLV encoding" % typ)  # noqa: E501
            n = 0
            while val:
                s = chb(val & 0xff) + s
                val >>= 8
                n += 1
            val = n
        else:
            type_val |= 0x8000
        return struct.pack("!HH", type_val, val) + s

    @staticmethod
    def num2type(typ, enc, proto=0):
        if proto == PROTO_ISAKMP:
            val = ISAKMPTransformNum.get(typ, (typ, {}))
        elif proto == PROTO_IPSEC_ESP:
            val = IPSECTransformNum.get(typ, (typ, {}))
        else:
            val = (typ, {})
        enc = val[1].get(enc, enc)
        return (val[0], enc)

    def _get_proto(self, pkt):
        # Ugh
        cur = pkt
        while cur and getattr(cur, "proto", None) is None:
            cur = cur.parent or cur.underlayer
        if cur is None:
            return PROTO_ISAKMP
        return cur.proto

    def i2m(self, pkt, i):
        if i is None:
            return b""
        proto = self._get_proto(pkt)
        i = [ISAKMPTransformSetField.type2num(e, proto=proto) for e in i]
        return b"".join(i)

    def m2i(self, pkt, m):
        # I try to ensure that we don't read off the end of our packet based
        # on bad length fields we're provided in the packet. There are still
        # conditions where struct.unpack() may not get enough packet data, but
        # worst case that should result in broken attributes (which would
        # be expected). (wam)
        lst = []
        proto = self._get_proto(pkt)
        while len(m) >= 4:
            trans_type, = struct.unpack("!H", m[:2])
            is_tlv = not (trans_type & 0x8000)
            if is_tlv:
                # We should probably check to make sure the attribute type we
                # are looking at is allowed to have a TLV format and issue a
                # warning if we're given an TLV on a basic attribute.
                value_len, = struct.unpack("!H", m[2:4])
                if value_len + 4 > len(m):
                    warning("Bad length for ISAKMP transform type=%#6x" % trans_type)  # noqa: E501
                value = m[4:4 + value_len]
                value = reduce(lambda x, y: (x << 8) | y, struct.unpack("!%s" % ("B" * len(value),), value), 0)  # noqa: E501
            else:
                trans_type &= 0x7fff
                value_len = 0
                value, = struct.unpack("!H", m[2:4])
            m = m[4 + value_len:]
            lst.append(ISAKMPTransformSetField.num2type(trans_type, value, proto=proto))
        if len(m) > 0:
            warning("Extra bytes after ISAKMP transform dissection [%r]" % m)
        return lst


ISAKMP_payload_type = {
    0: "None",
    1: "SA",
    2: "Proposal",
    3: "Transform",
    4: "KE",
    5: "ID",
    6: "CERT",
    7: "CR",
    8: "Hash",
    9: "SIG",
    10: "Nonce",
    11: "Notification",
    12: "Delete",
    13: "VendorID",
}

ISAKMP_exchange_type = {
    0: "None",
    1: "base",
    2: "identity protection",
    3: "authentication only",
    4: "aggressive",
    5: "informational",
    32: "quick mode",
}

# https://www.iana.org/assignments/isakmp-registry/isakmp-registry.xhtml#isakmp-registry-3
# IPSEC Security Protocol Identifiers
ISAKMP_protos = {
    1: "ISAKMP",
    2: "IPSEC_AH",
    3: "IPSEC_ESP",
    4: "IPCOMP",
    5: "GIGABEAM_RADIO"
}

ISAKMP_doi = {
    0: "ISAKMP",
    1: "IPSEC",
}


class _ISAKMP_class(Packet):
    def default_payload_class(self, payload):
        if self.next_payload == 0:
            return conf.raw_layer
        return ISAKMP_payload

# -- ISAKMP


class ISAKMP(_ISAKMP_class):  # rfc2408
    name = "ISAKMP"
    fields_desc = [
        XStrFixedLenField("init_cookie", "", 8),
        XStrFixedLenField("resp_cookie", "", 8),
        ByteEnumField("next_payload", 0, ISAKMP_payload_type),
        XByteField("version", 0x10),
        ByteEnumField("exch_type", 0, ISAKMP_exchange_type),
        FlagsField("flags", 0, 8, ["encryption", "commit", "auth_only"]),
        IntField("id", 0),
        IntField("length", None)
    ]

    def guess_payload_class(self, payload):
        if self.flags & 1:
            return conf.raw_layer
        return _ISAKMP_class.guess_payload_class(self, payload)

    def answers(self, other):
        if isinstance(other, ISAKMP):
            if other.init_cookie == self.init_cookie:
                return 1
        return 0

    def post_build(self, p, pay):
        p += pay
        if self.length is None:
            p = p[:24] + struct.pack("!I", len(p)) + p[28:]
        return p


# -- ISAKMP payloads

class ISAKMP_payload(_ISAKMP_class):
    name = "ISAKMP payload"
    show_indent = 0
    fields_desc = [
        ByteEnumField("next_payload", None, ISAKMP_payload_type),
        ByteField("res", 0),
        ShortField("length", None),
        XStrLenField("load", "", length_from=lambda x:x.length - 4),
    ]

    def post_build(self, pkt, pay):
        if self.length is None:
            pkt = pkt[:2] + struct.pack("!H", len(pkt)) + pkt[4:]
        return pkt + pay


class ISAKMP_payload_Transform(ISAKMP_payload):
    name = "IKE Transform"
    deprecated_fields = {
        "num": ("transform_count", ("2.5.0")),
        "id": ("transform_id", ("2.5.0")),
    }
    fields_desc = ISAKMP_payload.fields_desc[:3] + [
        ByteField("transform_count", None),
        ByteEnumField("transform_id", 1, {1: "KEY_IKE"}),
        ShortField("res2", 0),
        ISAKMPTransformSetField("transforms", None, length_from=lambda x: x.length - 8)  # noqa: E501
        #        XIntField("enc",0x80010005L),
        #        XIntField("hash",0x80020002L),
        #        XIntField("auth",0x80030001L),
        #        XIntField("group",0x80040002L),
        #        XIntField("life_type",0x800b0001L),
        #        XIntField("durationh",0x000c0004L),
        #        XIntField("durationl",0x00007080L),
    ]


# https://tools.ietf.org/html/rfc2408#section-3.5
class ISAKMP_payload_Proposal(ISAKMP_payload):
    name = "IKE proposal"
    fields_desc = ISAKMP_payload.fields_desc[:3] + [
        ByteField("proposal", 1),
        ByteEnumField("proto", 1, ISAKMP_protos),
        FieldLenField("SPIsize", None, "SPI", "B"),
        ByteField("trans_nb", None),
        StrLenField("SPI", "", length_from=lambda x: x.SPIsize),
        PacketLenField("trans", conf.raw_layer(), ISAKMP_payload_Transform, length_from=lambda x: x.length - 8),  # noqa: E501
    ]


# VendorID: https://www.rfc-editor.org/rfc/rfc2408#section-3.16

# packet-isakmp.c from wireshark
ISAKMP_VENDOR_IDS = {
    b"\x09\x00\x26\x89\xdf\xd6\xb7\x12": "XAUTH",
    b"\xaf\xca\xd7\x13h\xa1\xf1\xc9k\x86\x96\xfcwW\x01\x00": "RFC 3706 DPD",
    b"@H\xb7\xd5n\xbc\xe8\x85%\xe7\xde\x7f\x00\xd6\xc2\xd3\x80": "Cisco Fragmentation",
    b"J\x13\x1c\x81\x07\x03XE\\W(\xf2\x0e\x95E/": "RFC 3947 Negotiation of NAT-Transversal",  # noqa: E501
    b"\x90\xcb\x80\x91>\xbbin\x08c\x81\xb5\xecB{\x1f": "draft-ietf-ipsec-nat-t-ike-02",
}


class ISAKMP_payload_VendorID(ISAKMP_payload):
    name = "ISAKMP Vendor ID"
    fields_desc = ISAKMP_payload.fields_desc[:3] + [
        StrLenEnumField("VendorID", b"",
                        ISAKMP_VENDOR_IDS,
                        length_from=lambda x: x.length - 4)
    ]


class ISAKMP_payload_SA(ISAKMP_payload):
    name = "ISAKMP SA"
    fields_desc = ISAKMP_payload.fields_desc[:3] + [
        IntEnumField("doi", 1, ISAKMP_doi),
        IntEnumField("situation", 1, {1: "identity"}),
        PacketLenField("prop", conf.raw_layer(), ISAKMP_payload_Proposal, length_from=lambda x: x.length - 12),  # noqa: E501
    ]


class ISAKMP_payload_Nonce(ISAKMP_payload):
    name = "ISAKMP Nonce"
    deprecated_fields = {"load": ("nonce", "2.6.2")}
    fields_desc = ISAKMP_payload.fields_desc[:3] + [
        StrLenField("nonce", "", length_from=lambda x: x.length - 4)
    ]


class ISAKMP_payload_KE(ISAKMP_payload):
    name = "ISAKMP Key Exchange"
    deprecated_fields = {"load": ("ke", "2.6.2")}
    fields_desc = ISAKMP_payload.fields_desc[:3] + [
        StrLenField("ke", "", length_from=lambda x: x.length - 4)
    ]


class ISAKMP_payload_ID(ISAKMP_payload):
    name = "ISAKMP Identification"
    fields_desc = ISAKMP_payload.fields_desc[:3] + [
        ByteEnumField("IDtype", 1, {
            # Beware, apparently in-the-wild the values used
            # appear to be the ones from IKEv2 (RFC4306 sect 3.5)
            # and not ISAKMP (RFC2408 sect A.4)
            1: "IPv4_addr",
            11: "Key"
        }),
        ByteEnumField("ProtoID", 0, {0: "Unused"}),
        ShortEnumField("Port", 0, {0: "Unused"}),
        MultipleTypeField(
            [
                (IPField("IdentData", "127.0.0.1"),
                 lambda pkt: pkt.IDtype == 1),
            ],
            StrLenField("IdentData", "", length_from=lambda x: x.length - 8),
        )
    ]


class ISAKMP_payload_Hash(ISAKMP_payload):
    name = "ISAKMP Hash"
    deprecated_fields = {"load": ("hash", "2.6.2")}
    fields_desc = ISAKMP_payload.fields_desc[:3] + [
        StrLenField("hash", "", length_from=lambda x: x.length - 4)
    ]


class ISAKMP_payload_SIG(ISAKMP_payload):
    name = "ISAKMP Signature"
    deprecated_fields = {"load": ("sig", "2.6.2")}
    fields_desc = ISAKMP_payload.fields_desc[:3] + [
        StrLenField("sig", "", length_from=lambda x: x.length - 4)
    ]


NotifyMessageType = {
    1: "INVALID-PAYLOAD-TYPE",
    2: "DOI-NOT-SUPPORTED",
    3: "SITUATION-NOT-SUPPORTED",
    4: "INVALID-COOKIE",
    5: "INVALID-MAJOR-VERSION",
    6: "INVALID-MINOR-VERSION",
    7: "INVALID-EXCHANGE-TYPE",
    8: "INVALID-FLAGS",
    9: "INVALID-MESSAGE-ID",
    10: "INVALID-PROTOCOL-ID",
    11: "INVALID-SPI",
    12: "INVALID-TRANSFORM-ID",
    13: "ATTRIBUTES-NOT-SUPPORTED",
    14: "NO-PROPOSAL-CHOSEN",
    15: "BAD-PROPOSAL-SYNTAX",
    16: "PAYLOAD-MALFORMED",
    17: "INVALID-KEY-INFORMATION",
    18: "INVALID-ID-INFORMATION",
    19: "INVALID-CERT-ENCODING",
    20: "INVALID-CERTIFICATE",
    21: "CERT-TYPE-UNSUPPORTED",
    22: "INVALID-CERT-AUTHORITY",
    23: "INVALID-HASH-INFORMATION",
    24: "AUTHENTICATION-FAILED",
    25: "INVALID-SIGNATURE",
    26: "ADDRESS-NOTIFICATION",
    27: "NOTIFY-SA-LIFETIME",
    28: "CERTIFICATE-UNAVAILABLE",
    29: "UNSUPPORTED-EXCHANGE-TYPE",
    30: "UNEQUAL-PAYLOAD-LENGTHS",
    16384: "CONNECTED",
    # RFC 3706
    36136: "R-U-THERE",
    36137: "R-U-THERE-ACK",
}


class ISAKMP_payload_Notify(ISAKMP_payload):
    name = "ISAKMP Notify (Notification)"
    fields_desc = ISAKMP_payload.fields_desc[:3] + [
        IntEnumField("doi", 0, ISAKMP_doi),
        ByteEnumField("proto", 1, ISAKMP_protos),
        FieldLenField("SPIsize", None, "SPI", "B"),
        ShortEnumField("notify_msg_type", None, NotifyMessageType),
        StrLenField("SPI", "", length_from=lambda x: x.SPIsize),
        StrLenField("notify_data", "",
                    length_from=lambda x: x.length - x.SPIsize - 12)
    ]


class ISAKMP_payload_Delete(ISAKMP_payload):
    name = "ISAKMP Delete"
    fields_desc = ISAKMP_payload.fields_desc[:3] + [
        IntEnumField("doi", 0, ISAKMP_doi),
        ByteEnumField("proto", 1, ISAKMP_protos),
        FieldLenField("SPIsize", None, length_of="SPIs", fmt="B",
                      adjust=lambda pkt, x: x and x // len(pkt.SPIs)),
        FieldLenField("SPIcount", None, count_of="SPIs", fmt="H"),
        FieldListField("SPIs", [],
                       StrLenField("", "", length_from=lambda pkt: pkt.SPIsize),
                       count_from=lambda pkt: pkt.SPIcount),
    ]


bind_bottom_up(UDP, ISAKMP, dport=500)
bind_bottom_up(UDP, ISAKMP, sport=500)
bind_top_down(UDP, ISAKMP, dport=500, sport=500)

bind_bottom_up(NON_ESP, ISAKMP)

# Add bindings
bind_top_down(_ISAKMP_class, ISAKMP_payload, next_payload=0)
bind_layers(_ISAKMP_class, ISAKMP_payload_SA, next_payload=1)
bind_layers(_ISAKMP_class, ISAKMP_payload_Proposal, next_payload=2)
bind_layers(_ISAKMP_class, ISAKMP_payload_Transform, next_payload=3)
bind_layers(_ISAKMP_class, ISAKMP_payload_KE, next_payload=4)
bind_layers(_ISAKMP_class, ISAKMP_payload_ID, next_payload=5)
# bind_layers(_ISAKMP_class, ISAKMP_payload_CERT, next_payload=6)
# bind_layers(_ISAKMP_class, ISAKMP_payload_CR, next_payload=7)
bind_layers(_ISAKMP_class, ISAKMP_payload_Hash, next_payload=8)
bind_layers(_ISAKMP_class, ISAKMP_payload_SIG, next_payload=9)
bind_layers(_ISAKMP_class, ISAKMP_payload_Nonce, next_payload=10)
bind_layers(_ISAKMP_class, ISAKMP_payload_Notify, next_payload=11)
bind_layers(_ISAKMP_class, ISAKMP_payload_Delete, next_payload=12)
bind_layers(_ISAKMP_class, ISAKMP_payload_VendorID, next_payload=13)


def ikescan(ip):
    """Sends/receives a ISAMPK payload SA with payload proposal"""
    pkt = IP(dst=ip)
    pkt /= UDP()
    pkt /= ISAKMP(init_cookie=RandString(8), exch_type=2)
    pkt /= ISAKMP_payload_SA(prop=ISAKMP_payload_Proposal())
    return sr(pkt)



# SPDX-License-Identifier: GPL-2.0-only
# This file is part of Scapy
# See https://scapy.net/ for more information
# Copyright (C) Nils Weiss <nils@we155.de>
# Copyright (C) Markus Schroetter <project.m.schroetter@gmail.com>

# scapy.contrib.description = GMLAN Utilities
# scapy.contrib.status = loads

import time

from scapy.contrib.automotive import log_automotive

from scapy.contrib.automotive.gm.gmlan import GMLAN, GMLAN_SA, GMLAN_RD, \
    GMLAN_TD, GMLAN_PM, GMLAN_RMBA
from scapy.config import conf
from scapy.packet import Packet
from scapy.supersocket import SuperSocket
from scapy.contrib.isotp import ISOTPSocket
from scapy.utils import PeriodicSenderThread

from typing import (
    Optional,
    cast,
    Callable,
)

__all__ = ["GMLAN_TesterPresentSender", "GMLAN_InitDiagnostics",
           "GMLAN_GetSecurityAccess", "GMLAN_RequestDownload",
           "GMLAN_TransferData", "GMLAN_TransferPayload",
           "GMLAN_ReadMemoryByAddress", "GMLAN_BroadcastSocket"]

log_automotive.info("\"conf.contribs['GMLAN']"
                    "['treat-response-pending-as-answer']\" set to True). This "
                    "is required by the GMLAN-Utils module to operate "
                    "correctly.")
try:
    conf.contribs['GMLAN']['treat-response-pending-as-answer'] = False
except KeyError:
    conf.contribs['GMLAN'] = {'treat-response-pending-as-answer': False}


# Helper function
def _check_response(resp):
    # type: (Optional[Packet]) -> bool
    if resp is None:
        log_automotive.debug("Timeout.")
        return False
    log_automotive.debug("%s", repr(resp))
    return resp.service != 0x7f  # NegativeResponse


class GMLAN_TesterPresentSender(PeriodicSenderThread):

    def __init__(self, sock, pkt=GMLAN(service="TesterPresent"), interval=2):
        # type: (SuperSocket, Packet, int) -> None
        """ Thread to send GMLAN TesterPresent packets periodically

        :param sock: socket where packet is sent periodically
        :param pkt: packet to send
        :param interval: interval between two packets
        """
        PeriodicSenderThread.__init__(self, sock, pkt, interval)

    def run(self):
        # type: () -> None
        while not self._stopped.is_set() and not self._socket.closed:
            for p in self._pkts:
                self._socket.sr1(p, verbose=False, timeout=0.1)
                self._stopped.wait(timeout=self._interval)
                if self._stopped.is_set() or self._socket.closed:
                    break


def GMLAN_InitDiagnostics(
        sock,  # type: SuperSocket
        broadcast_socket=None,  # type: Optional[SuperSocket]
        timeout=1,  # type: int
        retry=0,  # type: int
        unittest=False  # type: bool
):
    # type: (...) -> bool
    """ Send messages to put an ECU into diagnostic/programming state.

    :param sock: socket for communication.
    :param broadcast_socket: socket for broadcasting. If provided some message
                             will be sent as broadcast. Recommended when used
                             on a network with several ECUs.
    :param timeout: timeout for sending, receiving or sniffing packages.
    :param retry: number of retries in case of failure.
    :param unittest: disable delays
    :return: True on success else False
    """

    # Helper function
    def _send_and_check_response(sock, req, timeout):
        # type: (SuperSocket, Packet, int) -> bool
        log_automotive.debug("Sending %s", repr(req))
        resp = sock.sr1(req, timeout=timeout, verbose=False)
        return _check_response(resp)

    retry = abs(retry)

    while retry >= 0:
        retry -= 1

        # DisableNormalCommunication
        p = GMLAN(service="DisableNormalCommunication")
        if broadcast_socket is None:
            if not _send_and_check_response(sock, p, timeout):
                continue
        else:
            log_automotive.debug("Sending %s as broadcast", repr(p))
            broadcast_socket.send(p)

        if not unittest:
            time.sleep(0.05)

        # ReportProgrammedState
        p = GMLAN(service="ReportProgrammingState")
        if not _send_and_check_response(sock, p, timeout):
            continue
        # ProgrammingMode requestProgramming
        p = GMLAN() / GMLAN_PM(subfunction="requestProgrammingMode")
        if not _send_and_check_response(sock, p, timeout):
            continue

        if not unittest:
            time.sleep(0.05)

        # InitiateProgramming enableProgramming
        # No response expected
        p = GMLAN() / GMLAN_PM(subfunction="enableProgrammingMode")
        log_automotive.debug("Sending %s", repr(p))
        sock.sr1(p, timeout=0.001, verbose=False)
        return True
    return False


def GMLAN_GetSecurityAccess(
        sock,  # type: SuperSocket
        key_function,  # type: Callable[[int], int]
        level=1,  # type: int
        timeout=None,  # type: Optional[int]
        retry=0,  # type: int
        unittest=False  # type: bool
):
    # type: (...) -> bool
    """ Authenticate on ECU. Implements Seey-Key procedure.

    :param sock: socket to send the message on.
    :param key_function: function implementing the key algorithm.
    :param level: level of access
    :param timeout: timeout for sending, receiving or sniffing packages.
    :param retry: number of retries in case of failure.
    :param unittest: disable internal delays
    :return: True on success.
    """
    retry = abs(retry)

    if key_function is None:
        return False

    if level % 2 == 0:
        log_automotive.warning("Parameter Error: Level must be an odd number.")
        return False

    while retry >= 0:
        retry -= 1

        request = GMLAN() / GMLAN_SA(subfunction=level)
        log_automotive.debug("Requesting seed..")
        resp = sock.sr1(request, timeout=timeout, verbose=False)
        if not _check_response(resp):
            if resp is not None and resp.returnCode == 0x37 and retry:
                log_automotive.debug("RequiredTimeDelayNotExpired. Wait 10s.")
                if not unittest:
                    time.sleep(10)
            log_automotive.debug("Negative Response.")
            continue

        seed = cast(Packet, resp).securitySeed
        if seed == 0:
            log_automotive.debug("ECU security already unlocked. (seed is 0x0000)")
            return True

        keypkt = GMLAN() / GMLAN_SA(subfunction=level + 1,
                                    securityKey=key_function(seed))
        log_automotive.debug("Responding with key..")
        resp = sock.sr1(keypkt, timeout=timeout, verbose=False)
        if resp is None:
            log_automotive.debug("Timeout.")
            continue
        log_automotive.debug("%s", repr(resp))
        if resp.service == 0x67:
            log_automotive.debug("SecurityAccess granted.")
            return True
        # Invalid Key
        elif resp.service == 0x7f and resp.returnCode == 0x35:
            log_automotive.debug("Key invalid")
            continue

    return False


def GMLAN_RequestDownload(sock, length, timeout=None, retry=0):
    # type: (SuperSocket, int, Optional[int], int) -> bool
    """ Send RequestDownload message.

        Usually used before calling TransferData.

    :param sock: socket to send the message on.
    :param length: value for the message's parameter 'unCompressedMemorySize'.
    :param timeout: timeout for sending, receiving or sniffing packages.
    :param retry: number of retries in case of failure.
    :return: True on success
    """
    retry = abs(retry)

    while retry >= 0:
        # RequestDownload
        pkt = GMLAN() / GMLAN_RD(memorySize=length)
        resp = sock.sr1(pkt, timeout=timeout, verbose=False)
        if _check_response(resp):
            return True
        retry -= 1
        if retry >= 0:
            log_automotive.debug("Retrying..")
    return False


def GMLAN_TransferData(
        sock,  # type: SuperSocket
        addr,  # type: int
        payload,  # type: bytes
        maxmsglen=None,  # type: Optional[int]
        timeout=None,  # type: Optional[int]
        retry=0  # type: int
):
    # type: (...) -> bool
    """ Send TransferData message.

    Usually used after calling RequestDownload.

    :param sock: socket to send the message on.
    :param addr: destination memory address on the ECU.
    :param payload: data to be sent.
    :param maxmsglen: maximum length of a single iso-tp message.
                      default: maximum length
    :param timeout: timeout for sending, receiving or sniffing packages.
    :param retry: number of retries in case of failure.
    :return: True on success.
    """
    retry = abs(retry)
    startretry = retry

    scheme = conf.contribs['GMLAN']['GMLAN_ECU_AddressingScheme']
    if addr < 0 or addr >= 2 ** (8 * scheme):
        log_automotive.warning("Error: Invalid address %s for scheme %s",
                               hex(addr), str(scheme))
        return False

    # max size of dataRecord according to gmlan protocol
    if maxmsglen is None or maxmsglen <= 0 or maxmsglen > (4093 - scheme):
        maxmsglen = (4093 - scheme)

    maxmsglen = cast(int, maxmsglen)

    for i in range(0, len(payload), maxmsglen):
        retry = startretry
        while True:
            if len(payload[i:]) > maxmsglen:
                transdata = payload[i:i + maxmsglen]
            else:
                transdata = payload[i:]
            pkt = GMLAN() / GMLAN_TD(startingAddress=addr + i,
                                     dataRecord=transdata)
            resp = sock.sr1(pkt, timeout=timeout, verbose=False)
            if _check_response(resp):
                break
            retry -= 1
            if retry >= 0:
                log_automotive.debug("Retrying..")
            else:
                return False

    return True


def GMLAN_TransferPayload(
        sock,  # type: SuperSocket
        addr,  # type: int
        payload,  # type: bytes
        maxmsglen=None,  # type: Optional[int]
        timeout=None,  # type: Optional[int]
        retry=0  # type: int
):
    # type: (...) -> bool
    """ Send data by using GMLAN services.

    :param sock: socket to send the data on.
    :param addr: destination memory address on the ECU.
    :param payload: data to be sent.
    :param maxmsglen: maximum length of a single iso-tp message.
                      default: maximum length
    :param timeout: timeout for sending, receiving or sniffing packages.
    :param retry: number of retries in case of failure.
    :return: True on success.
    """
    if not GMLAN_RequestDownload(sock, len(payload), timeout=timeout,
                                 retry=retry):
        return False
    if not GMLAN_TransferData(sock, addr, payload, maxmsglen=maxmsglen,
                              timeout=timeout, retry=retry):
        return False
    return True


def GMLAN_ReadMemoryByAddress(
        sock,  # type: SuperSocket
        addr,  # type: int
        length,  # type: int
        timeout=None,  # type: Optional[int]
        retry=0  # type: int
):
    # type: (...) -> Optional[bytes]
    """ Read data from ECU memory.

    :param sock: socket to send the data on.
    :param addr: source memory address on the ECU.
    :param length: bytes to read.
    :param timeout: timeout for sending, receiving or sniffing packages.
    :param retry: number of retries in case of failure.
    :return: bytes red or None
    """
    retry = abs(retry)

    scheme = conf.contribs['GMLAN']['GMLAN_ECU_AddressingScheme']
    if addr < 0 or addr >= 2 ** (8 * scheme):
        log_automotive.warning("Error: Invalid address %s for scheme %s",
                               hex(addr), str(scheme))
        return None

    # max size of dataRecord according to gmlan protocol
    if length <= 0 or length > (4094 - scheme):
        log_automotive.warning("Error: Invalid length %s for scheme %s. "
                               "Choose between 0x1 and %s",
                               hex(length), str(scheme), hex(4094 - scheme))
        return None

    while retry >= 0:
        # RequestDownload
        pkt = GMLAN() / GMLAN_RMBA(memoryAddress=addr, memorySize=length)
        resp = sock.sr1(pkt, timeout=timeout, verbose=False)
        if _check_response(resp):
            return cast(Packet, resp).dataRecord
        retry -= 1
        if retry >= 0:
            log_automotive.debug("Retrying..")
    return None


def GMLAN_BroadcastSocket(interface):
    # type: (str) -> SuperSocket
    """ Returns a GMLAN broadcast socket using interface.

    :param interface: interface name
    :return: ISOTPSocket configured as GMLAN Broadcast Socket
    """
    return ISOTPSocket(interface, tx_id=0x101, rx_id=0x0, basecls=GMLAN,
                       ext_address=0xfe, padding=True)



# SPDX-License-Identifier: GPL-2.0-or-later
# This file is part of Scapy
# See https://scapy.net/ for more information
# Copyright (C) 2009 Adline Stephane <adline.stephane@gmail.com>
# Copyright     2018 Gabriel Potter <gabriel[]potter[]fr>

"""
Secure Neighbor Discovery (SEND) - RFC3971
"""

# scapy.contrib.description = Secure Neighbor Discovery (SEND) (ICMPv6)
# scapy.contrib.status = loads


from scapy.packet import Packet
from scapy.fields import BitField, ByteField, FieldLenField, PacketField, \
    PacketLenField, ShortField, StrFixedLenField, StrLenField, UTCTimeField
from scapy.layers.x509 import X509_SubjectPublicKeyInfo
from scapy.layers.inet6 import icmp6ndoptscls, _ICMPv6NDGuessPayload
from scapy.compat import chb
from scapy.volatile import RandBin


class ICMPv6NDOptNonce(_ICMPv6NDGuessPayload, Packet):
    name = "ICMPv6NDOptNonce"
    fields_desc = [ByteField("type", 14),
                   FieldLenField("len", None, length_of="nonce", fmt="B", adjust=lambda pkt, x: int(round((x + 2) / 8.))),  # noqa: E501
                   StrLenField("nonce", "", length_from=lambda pkt: pkt.len * 8 - 2)]  # noqa: E501


class ICMPv6NDOptTmstp(_ICMPv6NDGuessPayload, Packet):
    name = "ICMPv6NDOptTmstp"
    fields_desc = [ByteField("type", 13),
                   ByteField("len", 2),
                   BitField("reserved", 0, 48),
                   UTCTimeField("timestamp", None)]


class ICMPv6NDOptRsaSig(_ICMPv6NDGuessPayload, Packet):
    name = "ICMPv6NDOptRsaSig"
    fields_desc = [ByteField("type", 12),
                   FieldLenField("len", None, length_of="signature_pad", fmt="B", adjust=lambda pkt, x: (x + 20) // 8),  # noqa: E501
                   ShortField("reserved", 0),
                   StrFixedLenField("key_hash", "", length=16),
                   StrLenField("signature_pad", "", length_from=lambda pkt: pkt.len * 8 - 20)]  # noqa: E501


class CGA_Params(Packet):
    name = "CGA Parameters data structure"
    fields_desc = [StrFixedLenField("modifier", RandBin(size=16), length=16),
                   StrFixedLenField("subprefix", "", length=8),
                   ByteField("cc", 0),
                   PacketField("pubkey", X509_SubjectPublicKeyInfo(),
                               X509_SubjectPublicKeyInfo)]


class ICMPv6NDOptCGA(_ICMPv6NDGuessPayload, Packet):
    name = "ICMPv6NDOptCGA"
    fields_desc = [ByteField("type", 11),
                   FieldLenField("len", None, length_of="CGA_PARAMS", fmt="B", adjust=lambda pkt, x: (x + pkt.padlength + 4) // 8),  # noqa: E501
                   FieldLenField("padlength", 0, length_of="padding", fmt="B"),
                   ByteField("reserved", 0),
                   PacketLenField("CGA_PARAMS", "", CGA_Params, length_from=lambda pkt: pkt.len * 8 - pkt.padlength - 4),  # noqa: E501
                   StrLenField("padding", "", length_from=lambda pkt: pkt.padlength)]  # noqa: E501

    def post_build(self, p, pay):
        l_ = len(self.CGA_PARAMS)
        tmp_len = -(4 + l_) % 8  # Pad to 8 bytes
        p = p[:1] + chb((4 + l_ + tmp_len) // 8) + chb(tmp_len) + p[3:4 + l_]
        p += b"\x00" * tmp_len + pay
        return p


send_icmp6ndoptscls = {11: ICMPv6NDOptCGA,
                       12: ICMPv6NDOptRsaSig,
                       13: ICMPv6NDOptTmstp,
                       14: ICMPv6NDOptNonce
                       }
icmp6ndoptscls.update(send_icmp6ndoptscls)



# SPDX-License-Identifier: GPL-2.0-only
# This file is part of Scapy
# See https://scapy.net/ for more information
# Copyright (C) Philippe Biondi <phil@secdev.org>

"""
Scapy: create, send, sniff, dissect and manipulate network packets.

Usable either from an interactive console or as a Python library.
https://scapy.net
"""

import datetime
import os
import re
import subprocess

__all__ = [
    "VERSION",
    "__version__",
]

_SCAPY_PKG_DIR = os.path.dirname(__file__)


def _parse_tag(tag):
    # type: (str) -> str
    """
    Parse a tag from ``git describe`` into a version.

    Example::

        v2.3.2-346-g164a52c075c8 -> '2.3.2.dev346'
    """
    match = re.match('^v?(.+?)-(\\d+)-g[a-f0-9]+$', tag)
    if match:
        # remove the 'v' prefix and add a '.devN' suffix
        return '%s.dev%s' % (match.group(1), match.group(2))
    else:
        match = re.match('^v?([\\d\\.]+(rc\\d+)?)$', tag)
        if match:
            # tagged release version
            return '%s' % (match.group(1))
        else:
            raise ValueError('tag has invalid format')


def _version_from_git_archive():
    # type: () -> str
    """
    Rely on git archive "export-subst" git attribute.
    See 'man gitattributes' for more details.
    Note: describe is only supported with git >= 2.32.0,
    and the `tags=true` option with git >= 2.35.0 but we
    use it to workaround GH#3121.
    """
    git_archive_id = '$Format:%ct %(describe:tags=true)$'.split()
    tstamp = git_archive_id[0]
    if len(git_archive_id) > 1:
        tag = git_archive_id[1]
    else:
        # project is run in CI and has another %(describe)
        tag = ""

    if "Format" in tstamp:
        raise ValueError('not a git archive')

    if "describe" in tag:
        # git is too old!
        tag = ""
    if tag:
        # archived revision is tagged, use the tag
        return _parse_tag(tag)
    elif tstamp:
        # archived revision is not tagged, use the commit date
        d = datetime.datetime.fromtimestamp(int(tstamp), datetime.timezone.utc)
        return d.strftime('%Y.%m.%d')

    raise ValueError("invalid git archive format")


def _version_from_git_describe():
    # type: () -> str
    """
    Read the version from ``git describe``. It returns the latest tag with an
    optional suffix if the current directory is not exactly on the tag.

    Example::

        $ git describe --always
        v2.3.2-346-g164a52c075c8

    The tag prefix (``v``) and the git commit sha1 (``-g164a52c075c8``) are
    removed if present.

    If the current directory is not exactly on the tag, a ``.devN`` suffix is
    appended where N is the number of commits made after the last tag.

    Example::

        >>> _version_from_git_describe()
        '2.3.2.dev346'

    :raises CalledProcessError: if git is unavailable
    :return: Scapy's latest tag
    """
    if not os.path.isdir(os.path.join(os.path.dirname(_SCAPY_PKG_DIR), '.git')):  # noqa: E501
        raise ValueError('not in scapy git repo')

    def _git(cmd):
        # type: (str) -> str
        process = subprocess.Popen(
            cmd.split(),
            cwd=_SCAPY_PKG_DIR,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE
        )
        out, err = process.communicate()
        if process.returncode == 0:
            return out.decode().strip()
        else:
            raise subprocess.CalledProcessError(process.returncode, err)

    tag = _git("git describe --tags --always --long")
    if not tag.startswith("v"):
        # Upstream was not fetched
        commit = _git("git rev-list --tags --max-count=1")
        tag = _git("git describe --tags --always --long %s" % commit)
    return _parse_tag(tag)


def _version():
    # type: () -> str
    """Returns the Scapy version from multiple methods

    :return: the Scapy version
    """
    # Method 0: from external packaging
    try:
        # possibly forced by external packaging
        return os.environ['SCAPY_VERSION']
    except KeyError:
        pass

    # Method 1: from the VERSION file, included in sdist and wheels
    version_file = os.path.join(_SCAPY_PKG_DIR, 'VERSION')
    try:
        # file generated when running sdist
        with open(version_file, 'r') as fdsec:
            tag = fdsec.read()
        return tag
    except (FileNotFoundError, NotADirectoryError):
        pass

    # Method 2: from the archive tag, exported when using git archives
    try:
        return _version_from_git_archive()
    except ValueError:
        pass

    # Method 3: from git itself, used when Scapy was cloned
    try:
        return _version_from_git_describe()
    except Exception:
        pass

    # Fallback
    try:
        # last resort, use the modification date of __init__.py
        d = datetime.datetime.fromtimestamp(
            os.path.getmtime(__file__), datetime.timezone.utc
        )
        return d.strftime('%Y.%m.%d')
    except Exception:
        pass

    # all hope is lost
    return '0.0.0'


VERSION = __version__ = _version()

_tmp = re.search(r"([0-9]|\.[0-9])+", VERSION)
VERSION_MAIN = _tmp.group() if _tmp is not None else VERSION

if __name__ == "__main__":
    from scapy.main import interact
    interact()



# SPDX-License-Identifier: GPL-2.0-only
# This file is part of Scapy
# See https://scapy.net/ for more information
# Copyright (C) Guillaume Valadon <guillaume@valadon.net>

"""
Scapy *BSD native support - core
"""


import fcntl
import os
import socket
import struct

from scapy.arch.bpf.consts import BIOCSETF, BIOCSETIF
from scapy.arch.common import compile_filter, free_filter
from scapy.config import conf
from scapy.consts import LINUX
from scapy.error import Scapy_Exception
from scapy.interfaces import (
    InterfaceProvider,
    NetworkInterface,
    _GlobInterfaceType,
)

# re-export
from scapy.arch.bpf.pfroute import (  # noqa F403
    read_routes,
    read_routes6,
    _get_if_list,
)
from scapy.arch.common import get_if_raw_addr, read_nameservers  # noqa: F401

# Typing
from typing import (
    Dict,
    List,
    Tuple,
)

if LINUX:
    raise OSError("BPF conflicts with Linux")

# BPF specific functions


def get_dev_bpf():
    # type: () -> Tuple[int, int]
    """Returns an opened BPF file object"""

    # Get the first available BPF handle
    for bpf in range(256):
        try:
            fd = os.open("/dev/bpf%i" % bpf, os.O_RDWR)
            return (fd, bpf)
        except OSError as ex:
            if ex.errno == 13:  # Permission denied
                raise Scapy_Exception(
                    (
                        "Permission denied: could not open /dev/bpf%i. "
                        "Make sure to be running Scapy as root ! (sudo)"
                    )
                    % bpf
                )
            continue

    raise Scapy_Exception("No /dev/bpf handle is available !")


def attach_filter(fd, bpf_filter, iface):
    # type: (int, str, _GlobInterfaceType) -> None
    """Attach a BPF filter to the BPF file descriptor"""
    bp = compile_filter(bpf_filter, iface)
    # Assign the BPF program to the interface
    ret = fcntl.ioctl(fd, BIOCSETF, bp)
    if ret < 0:
        raise Scapy_Exception("Can't attach the BPF filter !")
    # Free
    free_filter(bp)


def in6_getifaddr():
    # type: () -> List[Tuple[str, int, str]]
    """
    Returns a list of 3-tuples of the form (addr, scope, iface) where
    'addr' is the address of scope 'scope' associated to the interface
    'iface'.

    This is the list of all addresses of all interfaces available on
    the system.
    """
    ifaces = _get_if_list()
    return [
        (ip["address"], ip["scope"], iface["name"])
        for iface in ifaces.values()
        for ip in iface["ips"]
        if ip["af_family"] == socket.AF_INET6
    ]


# Interface provider


class BPFInterfaceProvider(InterfaceProvider):
    name = "BPF"

    def _is_valid(self, dev):
        # type: (NetworkInterface) -> bool
        if not dev.flags & 0x1:  # not IFF_UP
            return False
        # Get a BPF handle
        try:
            fd = get_dev_bpf()[0]
        except Scapy_Exception:
            return True  # Can't check if available (non sudo?)
        if fd is None:
            raise Scapy_Exception("No /dev/bpf are available !")
        # Check if the interface can be used
        try:
            fcntl.ioctl(fd, BIOCSETIF, struct.pack("16s16x", dev.network_name.encode()))
        except IOError:
            return False
        else:
            return True
        finally:
            # Close the file descriptor
            os.close(fd)

    def load(self):
        # type: () -> Dict[str, NetworkInterface]
        data = {}
        for iface in _get_if_list().values():
            if_data = iface.copy()
            if_data.update(
                {
                    "network_name": iface["name"],
                    "description": iface["name"],
                    "ips": [x["address"] for x in iface["ips"]],
                }
            )
            data[iface["name"]] = NetworkInterface(self, if_data)
        return data


conf.ifaces.register_provider(BPFInterfaceProvider)



# SPDX-License-Identifier: GPL-2.0-only
# This file is part of Scapy
# See https://scapy.net/ for more information
# Copyright (C) Philippe Biondi <phil@secdev.org>

"""
GPRS (General Packet Radio Service) for mobile data communication.
"""

from scapy.fields import StrStopField
from scapy.packet import Packet, bind_layers
from scapy.layers.inet import IP


class GPRS(Packet):
    name = "GPRSdummy"
    fields_desc = [
        StrStopField("dummy", "", b"\x65\x00\x00", 1)
    ]


bind_layers(GPRS, IP,)



# SPDX-License-Identifier: GPL-2.0-or-later
# This file is part of Scapy
# See https://scapy.net/ for more information
# Copyright (C) Michael Farrell <micolous+git@gmail.com>
#
# scapy.contrib.description = Eddystone BLE proximity beacon
# scapy.contrib.status = loads
"""
scapy.contrib.eddystone - Google Eddystone Bluetooth LE proximity beacons.

The Eddystone specification can be found at:
https://github.com/google/eddystone/blob/master/protocol-specification.md

These beacons are used as building blocks for other systems:

* Google's Physical Web <https://google.github.io/physical-web/>
* RuuviTag <https://github.com/ruuvi/ruuvi-sensor-protocols>
* Waze Beacons <https://www.waze.com/beacons>

"""

from scapy.compat import orb
from scapy.fields import IntField, SignedByteField, StrField, BitField, \
    StrFixedLenField, ShortField, FixedPointField, ByteEnumField
from scapy.layers.bluetooth import EIR_Hdr, EIR_ServiceData16BitUUID, \
    EIR_CompleteList16BitServiceUUIDs, LowEnergyBeaconHelper
from scapy.packet import bind_layers, Packet

EDDYSTONE_UUID = 0xfeaa

EDDYSTONE_URL_SCHEMES = {
    0: b"http://www.",
    1: b"https://www.",
    2: b"http://",
    3: b"https://",
}

EDDYSTONE_URL_TABLE = {
    0: b".com/",
    1: b".org/",
    2: b".edu/",
    3: b".net/",
    4: b".info/",
    5: b".biz/",
    6: b".gov/",
    7: b".com",
    8: b".org",
    9: b".edu",
    10: b".net",
    11: b".info",
    12: b".biz",
    13: b".gov",
}


class EddystoneURLField(StrField):
    # https://github.com/google/eddystone/tree/master/eddystone-url#eddystone-url-http-url-encoding
    def i2m(self, pkt, x):
        if x is None:
            return b""

        o = bytearray()
        p = 0
        while p < len(x):
            c = orb(x[p])
            if c == 46:  # "."
                for k, v in EDDYSTONE_URL_TABLE.items():
                    if x.startswith(v, p):
                        o.append(k)
                        p += len(v) - 1
                        break
                else:
                    o.append(c)
            else:
                o.append(c)
            p += 1

        # Make the output immutable.
        return bytes(o)

    def m2i(self, pkt, x):
        if not x:
            return None

        o = bytearray()
        for c in x:
            i = orb(c)
            r = EDDYSTONE_URL_TABLE.get(i)
            if r is None:
                o.append(i)
            else:
                o.extend(r)
        return bytes(o)

    def any2i(self, pkt, x):
        if isinstance(x, str):
            x = x.encode("ascii")
        return x


class Eddystone_Frame(Packet, LowEnergyBeaconHelper):
    """
    The base Eddystone frame on which all Eddystone messages are built.

    https://github.com/google/eddystone/blob/master/protocol-specification.md
    """
    name = "Eddystone Frame"
    fields_desc = [
        BitField("type", None, 4),
        BitField("reserved", 0, 4),
    ]

    def build_eir(self):
        """Builds a list of EIR messages to wrap this frame."""

        return LowEnergyBeaconHelper.base_eir + [
            EIR_Hdr() / EIR_CompleteList16BitServiceUUIDs(svc_uuids=[
                EDDYSTONE_UUID]),
            EIR_Hdr() / EIR_ServiceData16BitUUID() / self
        ]


class Eddystone_UID(Packet):
    """
    An Eddystone type for transmitting a unique identifier.

    https://github.com/google/eddystone/tree/master/eddystone-uid
    """
    name = "Eddystone UID"
    fields_desc = [
        SignedByteField("tx_power", 0),
        StrFixedLenField("namespace", None, 10),
        StrFixedLenField("instance", None, 6),
        StrFixedLenField("reserved", None, 2),
    ]


class Eddystone_URL(Packet):
    """
    An Eddystone type for transmitting a URL (to a web page).

    https://github.com/google/eddystone/tree/master/eddystone-url
    """
    name = "Eddystone URL"
    fields_desc = [
        SignedByteField("tx_power", 0),
        ByteEnumField("url_scheme", 0, EDDYSTONE_URL_SCHEMES),
        EddystoneURLField("url", None),
    ]

    def to_url(self):
        return EDDYSTONE_URL_SCHEMES[self.url_scheme] + self.url

    @staticmethod
    def from_url(url):
        """Creates an Eddystone_Frame with a Eddystone_URL for a given URL."""
        url = url.encode('ascii')
        scheme = None
        for k, v in EDDYSTONE_URL_SCHEMES.items():
            if url.startswith(v):
                scheme = k
                url = url[len(v):]
                break
        else:
            raise Exception("URLs must start with EDDYSTONE_URL_SCHEMES")

        return Eddystone_Frame() / Eddystone_URL(
            url_scheme=scheme,
            url=url)


class Eddystone_TLM(Packet):
    """
    An Eddystone type for transmitting beacon telemetry information.

    https://github.com/google/eddystone/tree/master/eddystone-tlm
    """
    name = "Eddystone TLM"
    fields_desc = [
        ByteEnumField("version", None, {
            0: "unencrypted",
            1: "encrypted",
        }),
    ]


class Eddystone_TLM_Unencrypted(Packet):
    """
    A subtype of Eddystone-TLM for transmitting telemetry in unencrypted form.

    https://github.com/google/eddystone/blob/master/eddystone-tlm/tlm-plain.md
    """
    name = "Eddystone TLM (Unencrypted)"
    fields_desc = [
        ShortField("batt_mv", 0),
        FixedPointField("temperature", -128, 16, 8),
        IntField("adv_cnt", None),
        IntField("sec_cnt", None),
    ]


class Eddystone_TLM_Encrypted(Packet):
    """
    A subtype of Eddystone-TLM for transmitting telemetry in encrypted form.

    This implementation does not support decrypting this data.

    https://github.com/google/eddystone/blob/master/eddystone-tlm/tlm-encrypted.md
    """
    name = "Eddystone TLM (Encrypted)"
    fields_desc = [
        StrFixedLenField("etlm", None, 12),
        StrFixedLenField("salt", None, 2),
        StrFixedLenField("mic", None, 2),
    ]


class Eddystone_EID(Packet):
    """
    An Eddystone type for transmitting encrypted, ephemeral identifiers.

    This implementation does not support decrypting this data.

    https://github.com/google/eddystone/tree/master/eddystone-eid
    """
    name = "Eddystone EID"
    fields_desc = [
        SignedByteField("tx_power", 0),
        StrFixedLenField("eid", None, 8),
    ]


bind_layers(Eddystone_TLM, Eddystone_TLM_Unencrypted, version=0)
bind_layers(Eddystone_TLM, Eddystone_TLM_Encrypted, version=1)

bind_layers(Eddystone_Frame, Eddystone_UID, type=0)
bind_layers(Eddystone_Frame, Eddystone_URL, type=1)
bind_layers(Eddystone_Frame, Eddystone_TLM, type=2)
bind_layers(Eddystone_Frame, Eddystone_EID, type=3)

bind_layers(EIR_ServiceData16BitUUID, Eddystone_Frame, svc_uuid=EDDYSTONE_UUID)



# SPDX-License-Identifier: GPL-2.0-only
# This file is part of Scapy
# See https://scapy.net/ for more information

"""
Wrap Scapy's shell in pyannotate.
"""

import os
import sys
sys.path.insert(0, os.path.abspath('../../'))

from pyannotate_runtime import collect_types  # noqa: E402
from scapy.main import interact  # noqa: E402

collect_types.init_types_collection()
with collect_types.collect():
    interact()

collect_types.dump_stats("pyannotate_results_main")



# SPDX-License-Identifier: GPL-2.0-only
# This file is part of Scapy
# See https://scapy.net/ for more information
# Copyright (C) Nils Weiss <nils@we155.de>

# scapy.contrib.description = Keyword Protocol 2000 (KWP2000) / ISO 14230
# scapy.contrib.status = loads

import struct

from scapy.fields import (
    BitField,
    ByteEnumField,
    ByteField,
    ConditionalField,
    MayEnd,
    ObservableDict,
    StrField,
    X3BytesField,
    XByteEnumField,
    XByteField,
    XShortEnumField,
)
from scapy.packet import Packet, bind_layers, NoPayload
from scapy.config import conf
from scapy.error import log_loading
from scapy.utils import PeriodicSenderThread
from scapy.plist import _PacketIterable
from scapy.contrib.isotp import ISOTP

from typing import (
    Dict,
    Any,
)


try:
    if conf.contribs['KWP']['treat-response-pending-as-answer']:
        pass
except KeyError:
    log_loading.info("Specify \"conf.contribs['KWP'] = "
                     "{'treat-response-pending-as-answer': True}\" to treat "
                     "a negative response 'requestCorrectlyReceived-"
                     "ResponsePending' as answer of a request. \n"
                     "The default value is False.")
    conf.contribs['KWP'] = {'treat-response-pending-as-answer': False}


class KWP(ISOTP):
    services = ObservableDict(
        {0x10: 'StartDiagnosticSession',
         0x11: 'ECUReset',
         0x14: 'ClearDiagnosticInformation',
         0x17: 'ReadStatusOfDiagnosticTroubleCodes',
         0x18: 'ReadDiagnosticTroubleCodesByStatus',
         0x1A: 'ReadECUIdentification',
         0x21: 'ReadDataByLocalIdentifier',
         0x22: 'ReadDataByIdentifier',
         0x23: 'ReadMemoryByAddress',
         0x27: 'SecurityAccess',
         0x28: 'DisableNormalMessageTransmission',
         0x29: 'EnableNormalMessageTransmission',
         0x2C: 'DynamicallyDefineLocalIdentifier',
         0x2E: 'WriteDataByIdentifier',
         0x30: 'InputOutputControlByLocalIdentifier',
         0x31: 'StartRoutineByLocalIdentifier',
         0x32: 'StopRoutineByLocalIdentifier',
         0x33: 'RequestRoutineResultsByLocalIdentifier',
         0x34: 'RequestDownload',
         0x35: 'RequestUpload',
         0x36: 'TransferData',
         0x37: 'RequestTransferExit',
         0x3B: 'WriteDataByLocalIdentifier',
         0x3D: 'WriteMemoryByAddress',
         0x3E: 'TesterPresent',
         0x85: 'ControlDTCSetting',
         0x86: 'ResponseOnEvent',
         0x50: 'StartDiagnosticSessionPositiveResponse',
         0x51: 'ECUResetPositiveResponse',
         0x54: 'ClearDiagnosticInformationPositiveResponse',
         0x57: 'ReadStatusOfDiagnosticTroubleCodesPositiveResponse',
         0x58: 'ReadDiagnosticTroubleCodesByStatusPositiveResponse',
         0x5A: 'ReadECUIdentificationPositiveResponse',
         0x61: 'ReadDataByLocalIdentifierPositiveResponse',
         0x62: 'ReadDataByIdentifierPositiveResponse',
         0x63: 'ReadMemoryByAddressPositiveResponse',
         0x67: 'SecurityAccessPositiveResponse',
         0x68: 'DisableNormalMessageTransmissionPositiveResponse',
         0x69: 'EnableNormalMessageTransmissionPositiveResponse',
         0x6C: 'DynamicallyDefineLocalIdentifierPositiveResponse',
         0x6E: 'WriteDataByIdentifierPositiveResponse',
         0x70: 'InputOutputControlByLocalIdentifierPositiveResponse',
         0x71: 'StartRoutineByLocalIdentifierPositiveResponse',
         0x72: 'StopRoutineByLocalIdentifierPositiveResponse',
         0x73: 'RequestRoutineResultsByLocalIdentifierPositiveResponse',
         0x74: 'RequestDownloadPositiveResponse',
         0x75: 'RequestUploadPositiveResponse',
         0x76: 'TransferDataPositiveResponse',
         0x77: 'RequestTransferExitPositiveResponse',
         0x7B: 'WriteDataByLocalIdentifierPositiveResponse',
         0x7D: 'WriteMemoryByAddressPositiveResponse',
         0x7E: 'TesterPresentPositiveResponse',
         0xC5: 'ControlDTCSettingPositiveResponse',
         0xC6: 'ResponseOnEventPositiveResponse',
         0x7f: 'NegativeResponse'})  # type: Dict[int, str]
    name = 'KWP'
    fields_desc = [
        XByteEnumField('service', 0, services)
    ]

    def answers(self, other):
        # type: (Packet) -> bool
        if not isinstance(other, type(self)):
            return False
        if self.service == 0x7f:
            return self.payload.answers(other)
        if self.service == (other.service + 0x40):
            if isinstance(self.payload, NoPayload) or \
                    isinstance(other.payload, NoPayload):
                return len(self) <= len(other)
            else:
                return self.payload.answers(other.payload)
        return False

    def hashret(self):
        # type: () -> bytes
        if self.service == 0x7f:
            return struct.pack('B', self.requestServiceId & ~0x40)
        else:
            return struct.pack('B', self.service & ~0x40)


# ########################SDS###################################
class KWP_SDS(Packet):
    diagnosticSessionTypes = ObservableDict({
        0x81: 'defaultSession',
        0x85: 'programmingSession',
        0x89: 'standBySession',
        0x90: 'EcuPassiveSession',
        0x92: 'extendedDiagnosticSession'})
    name = 'StartDiagnosticSession'
    fields_desc = [
        ByteEnumField('diagnosticSession', 0, diagnosticSessionTypes)
    ]


bind_layers(KWP, KWP_SDS, service=0x10)


class KWP_SDSPR(Packet):
    name = 'StartDiagnosticSessionPositiveResponse'
    fields_desc = [
        ByteEnumField('diagnosticSession', 0,
                      KWP_SDS.diagnosticSessionTypes),
    ]

    def answers(self, other):
        # type: (Packet) -> int
        return isinstance(other, KWP_SDS) and \
            other.diagnosticSession == self.diagnosticSession


bind_layers(KWP, KWP_SDSPR, service=0x50)


# ######################### KWP_ER ###################################
class KWP_ER(Packet):
    resetModes = {
        0x00: 'reserved',
        0x01: 'powerOnReset',
        0x82: 'nonvolatileMemoryReset'}
    name = 'ECUReset'
    fields_desc = [
        ByteEnumField('resetMode', 0, resetModes)
    ]


bind_layers(KWP, KWP_ER, service=0x11)


class KWP_ERPR(Packet):
    name = 'ECUResetPositiveResponse'

    def answers(self, other):
        # type: (Packet) -> int
        return isinstance(other, KWP_ER)


bind_layers(KWP, KWP_ERPR, service=0x51)


# ######################### KWP_SA ###################################
class KWP_SA(Packet):
    name = 'SecurityAccess'
    fields_desc = [
        ByteField('accessMode', 0),
        ConditionalField(StrField('key', b""),
                         lambda pkt: pkt.accessMode % 2 == 0)
    ]


bind_layers(KWP, KWP_SA, service=0x27)


class KWP_SAPR(Packet):
    name = 'SecurityAccessPositiveResponse'
    fields_desc = [
        ByteField('accessMode', 0),
        ConditionalField(StrField('seed', b""),
                         lambda pkt: pkt.accessMode % 2 == 1),
    ]

    def answers(self, other):
        # type: (Packet) -> int
        return isinstance(other, KWP_SA) \
            and other.accessMode == self.accessMode


bind_layers(KWP, KWP_SAPR, service=0x67)


# ######################### KWP_IOCBLI ###################################
class KWP_IOCBLI(Packet):
    name = 'InputOutputControlByLocalIdentifier'
    inputOutputControlParameters = {
        0x00: "Return Control to ECU",
        0x01: "Report Current State",
        0x04: "Reset to Default",
        0x05: "Freeze Current State",
        0x07: "Short Term Adjustment",
        0x08: "Long Term Adjustment"
    }
    fields_desc = [
        XByteField('localIdentifier', 0),
        XByteEnumField('inputOutputControlParameter', 0,
                       inputOutputControlParameters),
        StrField('controlState', b"", fmt="B")
    ]


bind_layers(KWP, KWP_IOCBLI, service=0x30)


class KWP_IOCBLIPR(Packet):
    name = 'InputOutputControlByLocalIdentifierPositiveResponse'
    fields_desc = [
        XByteField('localIdentifier', 0),
        XByteEnumField('inputOutputControlParameter', 0,
                       KWP_IOCBLI.inputOutputControlParameters),
        StrField('controlState', b"", fmt="B")
    ]

    def answers(self, other):
        # type: (Packet) -> int
        return isinstance(other, KWP_IOCBLI) \
            and other.localIdentifier == self.localIdentifier


bind_layers(KWP, KWP_IOCBLIPR, service=0x70)


# ######################### KWP_DNMT ###################################
class KWP_DNMT(Packet):
    responseTypes = {
        0x01: 'responseRequired',
        0x02: 'noResponse',
    }
    name = 'DisableNormalMessageTransmission'
    fields_desc = [
        ByteEnumField('responseRequired', 0, responseTypes)
    ]


bind_layers(KWP, KWP_DNMT, service=0x28)


class KWP_DNMTPR(Packet):
    name = 'DisableNormalMessageTransmissionPositiveResponse'

    def answers(self, other):
        # type: (Packet) -> int
        return isinstance(other, KWP_DNMT)


bind_layers(KWP, KWP_DNMTPR, service=0x68)


# ######################### KWP_ENMT ###################################
class KWP_ENMT(Packet):
    responseTypes = {
        0x01: 'responseRequired',
        0x02: 'noResponse',
    }
    name = 'EnableNormalMessageTransmission'
    fields_desc = [
        ByteEnumField('responseRequired', 1, responseTypes)
    ]


bind_layers(KWP, KWP_ENMT, service=0x29)


class KWP_ENMTPR(Packet):
    name = 'EnableNormalMessageTransmissionPositiveResponse'

    def answers(self, other):
        # type: (Packet) -> int
        return isinstance(other, KWP_DNMT)


bind_layers(KWP, KWP_ENMTPR, service=0x69)


# ######################### KWP_TP ###################################
class KWP_TP(Packet):
    responseTypes = {
        0x01: 'responseRequired',
        0x02: 'noResponse',
    }
    name = 'TesterPresent'
    fields_desc = [
        ByteEnumField('responseRequired', 1, responseTypes)
    ]


bind_layers(KWP, KWP_TP, service=0x3E)


class KWP_TPPR(Packet):
    name = 'TesterPresentPositiveResponse'

    def answers(self, other):
        # type: (Packet) -> int
        return isinstance(other, KWP_TP)


bind_layers(KWP, KWP_TPPR, service=0x7E)


# ######################### KWP_CDTCS ###################################
class KWP_CDTCS(Packet):
    responseTypes = {
        0x01: 'responseRequired',
        0x02: 'noResponse',
    }
    DTCGroups = {
        0x0000: 'allPowertrainDTCs',
        0x4000: 'allChassisDTCs',
        0x8000: 'allBodyDTCs',
        0xC000: 'allNetworkDTCs',
        0xFF00: 'allDTCs'
    }
    DTCSettingModes = {
        0: 'Reserved',
        1: 'on',
        2: 'off'
    }
    name = 'ControlDTCSetting'
    fields_desc = [
        ByteEnumField('responseRequired', 1, responseTypes),
        XShortEnumField('groupOfDTC', 0, DTCGroups),
        ByteEnumField('DTCSettingMode', 0, DTCSettingModes),
    ]


bind_layers(KWP, KWP_CDTCS, service=0x85)


class KWP_CDTCSPR(Packet):
    name = 'ControlDTCSettingPositiveResponse'

    def answers(self, other):
        # type: (Packet) -> int
        return isinstance(other, KWP_CDTCS)


bind_layers(KWP, KWP_CDTCSPR, service=0xC5)


# ######################### KWP_ROE ###################################
class KWP_ROE(Packet):
    responseTypes = {
        0x01: 'responseRequired',
        0x02: 'noResponse',
    }
    eventWindowTimes = {
        0x00: 'reserved',
        0x01: 'testerPresentRequired',
        0x02: 'infiniteTimeToResponse',
        0x80: 'noEventWindow'
    }
    eventTypes = {
        0x80: 'reportActivatedEvents',
        0x81: 'stopResponseOnEvent',
        0x82: 'onNewDTC',
        0x83: 'onTimerInterrupt',
        0x84: 'onChangeOfRecordValue',
        0xA0: 'onComparisonOfValues'
    }
    name = 'ResponseOnEvent'
    fields_desc = [
        ByteEnumField('responseRequired', 1, responseTypes),
        ByteEnumField('eventWindowTime', 0, eventWindowTimes),
        MayEnd(ByteEnumField('eventType', 0, eventTypes)),
        # XXX Is this MayEnd correct?
        ByteField('eventParameter', 0),
        ByteEnumField('serviceToRespond', 0, KWP.services),
        ByteField('serviceParameter', 0)
    ]


bind_layers(KWP, KWP_ROE, service=0x86)


class KWP_ROEPR(Packet):
    name = 'ResponseOnEventPositiveResponse'
    fields_desc = [
        ByteField("numberOfActivatedEvents", 0),
        MayEnd(ByteEnumField('eventWindowTime', 0, KWP_ROE.eventWindowTimes)),
        # XXX Is this MayEnd correct?
        ByteEnumField('eventType', 0, KWP_ROE.eventTypes),
    ]

    def answers(self, other):
        # type: (Packet) -> int
        return isinstance(other, KWP_ROE) \
            and other.eventType == self.eventType


bind_layers(KWP, KWP_ROEPR, service=0xC6)


# ######################### KWP_RDBLI ###################################
class KWP_RDBLI(Packet):
    localIdentifiers = ObservableDict({
        0xE0: "Development Data",
        0xE1: "ECU Serial Number",
        0xE2: "DBCom Data",
        0xE3: "Operating System Version",
        0xE4: "Ecu Reprogramming Identification",
        0xE5: "Vehicle Information",
        0xE6: "Flash Info 1",
        0xE7: "Flash Info 2",
        0xE8: "System Diagnostic general parameter data",
        0xE9: "System Diagnostic global parameter data",
        0xEA: "Ecu Configuration",
        0xEB: "Diagnostic Protocol Information"
    })
    name = 'ReadDataByLocalIdentifier'
    fields_desc = [
        XByteEnumField('recordLocalIdentifier', 0, localIdentifiers)
    ]


bind_layers(KWP, KWP_RDBLI, service=0x21)


class KWP_RDBLIPR(Packet):
    name = 'ReadDataByLocalIdentifierPositiveResponse'
    fields_desc = [
        XByteEnumField('recordLocalIdentifier', 0, KWP_RDBLI.localIdentifiers)
    ]

    def answers(self, other):
        # type: (Packet) -> int
        return isinstance(other, KWP_RDBLI) \
            and self.recordLocalIdentifier == other.recordLocalIdentifier


bind_layers(KWP, KWP_RDBLIPR, service=0x61)


# ######################### KWP_WDBLI ###################################
class KWP_WDBLI(Packet):
    name = 'WriteDataByLocalIdentifier'
    fields_desc = [
        XByteEnumField('recordLocalIdentifier', 0, KWP_RDBLI.localIdentifiers)
    ]


bind_layers(KWP, KWP_WDBLI, service=0x3B)


class KWP_WDBLIPR(Packet):
    name = 'WriteDataByLocalIdentifierPositiveResponse'
    fields_desc = [
        XByteEnumField('recordLocalIdentifier', 0, KWP_RDBLI.localIdentifiers)
    ]

    def answers(self, other):
        # type: (Packet) -> int
        return isinstance(other, KWP_WDBLI) \
            and self.recordLocalIdentifier == other.recordLocalIdentifier


bind_layers(KWP, KWP_WDBLIPR, service=0x7B)


# ######################### KWP_RDBI ###################################
class KWP_RDBI(Packet):
    dataIdentifiers = ObservableDict()
    name = 'ReadDataByIdentifier'
    fields_desc = [
        XShortEnumField('identifier', 0, dataIdentifiers)
    ]


bind_layers(KWP, KWP_RDBI, service=0x22)


class KWP_RDBIPR(Packet):
    name = 'ReadDataByIdentifierPositiveResponse'
    fields_desc = [
        XShortEnumField('identifier', 0, KWP_RDBI.dataIdentifiers),
    ]

    def answers(self, other):
        # type: (Packet) -> int
        return isinstance(other, KWP_RDBI) \
            and self.identifier == other.identifier


bind_layers(KWP, KWP_RDBIPR, service=0x62)


# ######################### KWP_RMBA ###################################
class KWP_RMBA(Packet):
    name = 'ReadMemoryByAddress'
    fields_desc = [
        X3BytesField('memoryAddress', 0),
        ByteField('memorySize', 0)
    ]


bind_layers(KWP, KWP_RMBA, service=0x23)


class KWP_RMBAPR(Packet):
    name = 'ReadMemoryByAddressPositiveResponse'
    fields_desc = [
        StrField('dataRecord', b"", fmt="B")
    ]

    def answers(self, other):
        # type: (Packet) -> int
        return isinstance(other, KWP_RMBA)


bind_layers(KWP, KWP_RMBAPR, service=0x63)


# ######################### KWP_DDLI ###################################
# TODO: Implement correct interpretation here,
#       instead of using just the dataRecord
class KWP_DDLI(Packet):
    name = 'DynamicallyDefineLocalIdentifier'
    definitionModes = {0x1: "defineByLocalIdentifier",
                       0x2: "defineByMemoryAddress",
                       0x3: "defineByIdentifier",
                       0x4: "clearDynamicallyDefinedLocalIdentifier"}
    fields_desc = [
        XByteField('dynamicallyDefineLocalIdentifier', 0),
        ByteEnumField('definitionMode', 0, definitionModes),
        StrField('dataRecord', b"", fmt="B")
    ]


bind_layers(KWP, KWP_DDLI, service=0x2C)


class KWP_DDLIPR(Packet):
    name = 'DynamicallyDefineLocalIdentifierPositiveResponse'
    fields_desc = [
        XByteField('dynamicallyDefineLocalIdentifier', 0)
    ]

    def answers(self, other):
        # type: (Packet) -> int
        return isinstance(other, KWP_DDLI) and \
            other.dynamicallyDefineLocalIdentifier == self.dynamicallyDefineLocalIdentifier  # noqa: E501


bind_layers(KWP, KWP_DDLIPR, service=0x6C)


# ######################### KWP_WDBI ###################################
class KWP_WDBI(Packet):
    name = 'WriteDataByIdentifier'
    fields_desc = [
        XShortEnumField('identifier', 0, KWP_RDBI.dataIdentifiers)
    ]


bind_layers(KWP, KWP_WDBI, service=0x2E)


class KWP_WDBIPR(Packet):
    name = 'WriteDataByIdentifierPositiveResponse'
    fields_desc = [
        XShortEnumField('identifier', 0, KWP_RDBI.dataIdentifiers),
    ]

    def answers(self, other):
        # type: (Packet) -> int
        return isinstance(other, KWP_WDBI) \
            and other.identifier == self.identifier


bind_layers(KWP, KWP_WDBIPR, service=0x6E)


# ######################### KWP_WMBA ###################################
class KWP_WMBA(Packet):
    name = 'WriteMemoryByAddress'
    fields_desc = [
        X3BytesField('memoryAddress', 0),
        ByteField('memorySize', 0),
        StrField('dataRecord', b'', fmt="B")
    ]


bind_layers(KWP, KWP_WMBA, service=0x3D)


class KWP_WMBAPR(Packet):
    name = 'WriteMemoryByAddressPositiveResponse'
    fields_desc = [
        X3BytesField('memoryAddress', 0)
    ]

    def answers(self, other):
        # type: (Packet) -> int
        return isinstance(other, KWP_WMBA) and \
            other.memoryAddress == self.memoryAddress


bind_layers(KWP, KWP_WMBAPR, service=0x7D)


# ######################### KWP_CDI ###################################
class KWP_CDI(Packet):
    DTCGroups = {
        0x0000: 'allPowertrainDTCs',
        0x4000: 'allChassisDTCs',
        0x8000: 'allBodyDTCs',
        0xC000: 'allNetworkDTCs',
        0xFF00: 'allDTCs'
    }
    name = 'ClearDiagnosticInformation'
    fields_desc = [
        XShortEnumField('groupOfDTC', 0, DTCGroups)
    ]


bind_layers(KWP, KWP_CDI, service=0x14)


class KWP_CDIPR(Packet):
    name = 'ClearDiagnosticInformationPositiveResponse'

    fields_desc = [
        XShortEnumField('groupOfDTC', 0, KWP_CDI.DTCGroups)
    ]

    def answers(self, other):
        # type: (Packet) -> int
        return isinstance(other, KWP_CDI) and \
            self.groupOfDTC == other.groupOfDTC


bind_layers(KWP, KWP_CDIPR, service=0x54)


# ######################### KWP_RSODTC ###################################
class KWP_RSODTC(Packet):
    name = 'ReadStatusOfDiagnosticTroubleCodes'
    fields_desc = [
        XShortEnumField('groupOfDTC', 0, KWP_CDI.DTCGroups)
    ]


bind_layers(KWP, KWP_RSODTC, service=0x17)


class KWP_RSODTCPR(Packet):
    name = 'ReadStatusOfDiagnosticTroubleCodesPositiveResponse'

    fields_desc = [
        ByteField('numberOfDTC', 0),
    ]

    def answers(self, other):
        # type: (Packet) -> int
        return isinstance(other, KWP_RSODTC)


bind_layers(KWP, KWP_RSODTCPR, service=0x57)


# ######################### KWP_RECUI ###################################
class KWP_RECUI(Packet):
    name = 'ReadECUIdentification'
    localIdentifiers = ObservableDict({
        0x86: "DCS ECU Identification",
        0x87: "DCX / MMC ECU Identification",
        0x88: "VIN (Original)",
        0x89: "Diagnostic Variant Code",
        0x90: "VIN (Current)",
        0x96: "Calibration Identification",
        0x97: "Calibration Verification Number",
        0x9A: "ECU Code Fingerprint",
        0x98: "ECU Data Fingerprint",
        0x9C: "ECU Code Software Identification",
        0x9D: "ECU Data Software Identification",
        0x9E: "ECU Boot Software Identification",
        0x9F: "ECU Boot Fingerprint"
    })
    fields_desc = [
        XByteEnumField('localIdentifier', 0, localIdentifiers)
    ]


bind_layers(KWP, KWP_RECUI, service=0x1A)


class KWP_RECUIPR(Packet):
    name = 'ReadECUIdentificationPositiveResponse'

    fields_desc = [
        XByteEnumField('localIdentifier', 0, KWP_RECUI.localIdentifiers)
    ]

    def answers(self, other):
        # type: (Packet) -> int
        return isinstance(other, KWP_RECUI) and \
            self.localIdentifier == other.localIdentifier


bind_layers(KWP, KWP_RECUIPR, service=0x5A)


# ######################### KWP_SRBLI ###################################
class KWP_SRBLI(Packet):
    routineLocalIdentifiers = ObservableDict({
        0xE0: "FlashEraseRoutine",
        0xE1: "FlashCheckRoutine",
        0xE2: "Tell-TaleRetentionStack",
        0xE3: "RequestDTCsFromShadowErrorMemory",
        0xE4: "RequestEnvironmentDataFromShadowErrorMemory",
        0xE5: "RequestEventInformation",
        0xE6: "RequestEventEnvironmentData",
        0xE7: "RequestSoftwareModuleInformation",
        0xE8: "ClearTell-TaleRetentionStack",
        0xE9: "ClearEventInformation"
    })
    name = 'StartRoutineByLocalIdentifier'
    fields_desc = [
        XByteEnumField('routineLocalIdentifier', 0, routineLocalIdentifiers)
    ]


bind_layers(KWP, KWP_SRBLI, service=0x31)


class KWP_SRBLIPR(Packet):
    name = 'StartRoutineByLocalIdentifierPositiveResponse'
    fields_desc = [
        XByteEnumField('routineLocalIdentifier', 0,
                       KWP_SRBLI.routineLocalIdentifiers)
    ]

    def answers(self, other):
        # type: (Packet) -> int
        return isinstance(other, KWP_SRBLI) \
            and other.routineLocalIdentifier == self.routineLocalIdentifier


bind_layers(KWP, KWP_SRBLIPR, service=0x71)


# ######################### KWP_STRBLI ###################################
class KWP_STRBLI(Packet):
    name = 'StopRoutineByLocalIdentifier'
    fields_desc = [
        XByteEnumField('routineLocalIdentifier', 0,
                       KWP_SRBLI.routineLocalIdentifiers)
    ]


bind_layers(KWP, KWP_STRBLI, service=0x32)


class KWP_STRBLIPR(Packet):
    name = 'StopRoutineByLocalIdentifierPositiveResponse'
    fields_desc = [
        XByteEnumField('routineLocalIdentifier', 0,
                       KWP_SRBLI.routineLocalIdentifiers)
    ]

    def answers(self, other):
        # type: (Packet) -> int
        return isinstance(other, KWP_STRBLI) \
            and other.routineLocalIdentifier == self.routineLocalIdentifier


bind_layers(KWP, KWP_STRBLIPR, service=0x72)


# ######################### KWP_RRRBLI ###################################
class KWP_RRRBLI(Packet):
    name = 'RequestRoutineResultsByLocalIdentifier'
    fields_desc = [
        XByteEnumField('routineLocalIdentifier', 0,
                       KWP_SRBLI.routineLocalIdentifiers)
    ]


bind_layers(KWP, KWP_RRRBLI, service=0x33)


class KWP_RRRBLIPR(Packet):
    name = 'RequestRoutineResultsByLocalIdentifierPositiveResponse'
    fields_desc = [
        XByteEnumField('routineLocalIdentifier', 0,
                       KWP_SRBLI.routineLocalIdentifiers)
    ]

    def answers(self, other):
        # type: (Packet) -> int
        return isinstance(other, KWP_RRRBLI) \
            and other.routineLocalIdentifier == self.routineLocalIdentifier


bind_layers(KWP, KWP_RRRBLIPR, service=0x73)


# ######################### KWP_RD ###################################
class KWP_RD(Packet):
    name = 'RequestDownload'
    fields_desc = [
        X3BytesField('memoryAddress', 0),
        BitField('compression', 0, 4),
        BitField('encryption', 0, 4),
        X3BytesField('uncompressedMemorySize', 0)
    ]


bind_layers(KWP, KWP_RD, service=0x34)


class KWP_RDPR(Packet):
    name = 'RequestDownloadPositiveResponse'
    fields_desc = [
        StrField('maxNumberOfBlockLength', b"", fmt="B"),
    ]

    def answers(self, other):
        # type: (Packet) -> int
        return isinstance(other, KWP_RD)


bind_layers(KWP, KWP_RDPR, service=0x74)


# ######################### KWP_RU ###################################
class KWP_RU(Packet):
    name = 'RequestUpload'
    fields_desc = [
        X3BytesField('memoryAddress', 0),
        BitField('compression', 0, 4),
        BitField('encryption', 0, 4),
        X3BytesField('uncompressedMemorySize', 0)
    ]


bind_layers(KWP, KWP_RU, service=0x35)


class KWP_RUPR(Packet):
    name = 'RequestUploadPositiveResponse'
    fields_desc = [
        StrField('maxNumberOfBlockLength', b"", fmt="B"),
    ]

    def answers(self, other):
        # type: (Packet) -> int
        return isinstance(other, KWP_RU)


bind_layers(KWP, KWP_RUPR, service=0x75)


# ######################### KWP_TD ###################################
class KWP_TD(Packet):
    name = 'TransferData'
    fields_desc = [
        ByteField('blockSequenceCounter', 0),
        StrField('transferDataRequestParameter', b"", fmt="B")
    ]


bind_layers(KWP, KWP_TD, service=0x36)


class KWP_TDPR(Packet):
    name = 'TransferDataPositiveResponse'
    fields_desc = [
        ByteField('blockSequenceCounter', 0),
        StrField('transferDataRequestParameter', b"", fmt="B")
    ]

    def answers(self, other):
        # type: (Packet) -> int
        return isinstance(other, KWP_TD) \
            and other.blockSequenceCounter == self.blockSequenceCounter


bind_layers(KWP, KWP_TDPR, service=0x76)


# ######################### KWP_RTE ###################################
class KWP_RTE(Packet):
    name = 'RequestTransferExit'
    fields_desc = [
        StrField('transferDataRequestParameter', b"", fmt="B")
    ]


bind_layers(KWP, KWP_RTE, service=0x37)


class KWP_RTEPR(Packet):
    name = 'RequestTransferExitPositiveResponse'
    fields_desc = [
        StrField('transferDataRequestParameter', b"", fmt="B")
    ]

    def answers(self, other):
        # type: (Packet) -> int
        return isinstance(other, KWP_RTE)


bind_layers(KWP, KWP_RTEPR, service=0x77)


# ######################### KWP_NR ###################################
class KWP_NR(Packet):
    negativeResponseCodes = {
        0x00: 'positiveResponse',
        0x10: 'generalReject',
        0x11: 'serviceNotSupported',
        0x12: 'subFunctionNotSupported-InvalidFormat',
        0x21: 'busyRepeatRequest',
        0x22: 'conditionsNotCorrect-RequestSequenceError',
        0x23: 'routineNotComplete',
        0x31: 'requestOutOfRange',
        0x33: 'securityAccessDenied-SecurityAccessRequested',
        0x35: 'invalidKey',
        0x36: 'exceedNumberOfAttempts',
        0x37: 'requiredTimeDelayNotExpired',
        0x40: 'downloadNotAccepted',
        0x50: 'uploadNotAccepted',
        0x71: 'transferSuspended',
        0x78: 'requestCorrectlyReceived-ResponsePending',
        0x80: 'subFunctionNotSupportedInActiveDiagnosticSession',
        0x9A: 'dataDecompressionFailed',
        0x9B: 'dataDecryptionFailed',
        0xA0: 'EcuNotResponding',
        0xA1: 'EcuAddressUnknown'
    }
    name = 'NegativeResponse'
    fields_desc = [
        MayEnd(XByteEnumField('requestServiceId', 0, KWP.services)),
        # XXX Is this MayEnd correct?
        ByteEnumField('negativeResponseCode', 0, negativeResponseCodes)
    ]

    def answers(self, other):
        # type: (Packet) -> int
        return self.requestServiceId == other.service and \
            (self.negativeResponseCode != 0x78 or
             conf.contribs['KWP']['treat-response-pending-as-answer'])


bind_layers(KWP, KWP_NR, service=0x7f)


# ##################################################################
# ######################## UTILS ###################################
# ##################################################################

class KWP_TesterPresentSender(PeriodicSenderThread):
    def __init__(self, sock, pkt=KWP() / KWP_TP(responseRequired=0x02),
                 interval=2):
        # type: (Any, _PacketIterable, float) -> None
        """ Thread that sends TesterPresent packets periodically

        :param sock: socket where packet is sent periodically
        :param pkt: packet to send
        :param interval: interval between two packets
        """
        PeriodicSenderThread.__init__(self, sock, pkt, interval)

    def run(self):
        # type: () -> None
        while not self._stopped.is_set():
            for p in self._pkts:
                self._socket.sr1(p, timeout=0.3, verbose=False)
                self._stopped.wait(timeout=self._interval)
                if self._stopped.is_set() or self._socket.closed:
                    break



# SPDX-License-Identifier: GPL-2.0-only
# This file is part of Scapy
# See https://scapy.net/ for more information
# Copyright (C) Tabea Spahn <tabea.spahn@e-mundo.de>

# scapy.contrib.status = skip

import struct

from scapy.config import conf
from scapy.contrib.automotive import log_automotive
from scapy.fields import StrLenField
from scapy.volatile import RandBin, RandNum


def get_max_cto():
    max_cto = conf.contribs['XCP']['MAX_CTO']
    if max_cto:
        return max_cto

    log_automotive.warning("Define conf.contribs['XCP']['MAX_CTO'].")
    raise KeyError("conf.contribs['XCP']['MAX_CTO'] not defined")


def get_max_dto():
    max_dto = conf.contribs['XCP']['MAX_DTO']
    if max_dto:
        return max_dto
    else:
        log_automotive.warning("Define conf.contribs['XCP']['MAX_DTO'].")
        raise KeyError("conf.contribs['XCP']['MAX_DTO'] not defined")


def get_ag():
    address_granularity = conf.contribs['XCP']['Address_Granularity_Byte']
    if address_granularity and address_granularity in [1, 2, 4]:
        return address_granularity
    else:
        log_automotive.warning(
            "Define conf.contribs['XCP']['Address_Granularity_Byte']."
            "Assign either 1, 2 or 4")
        return 1


# With TIMESTAMP_MODE and TIMESTAMP_TICKS at GET_DAQ_RESOLUTION_INFO,
# the slave informs the master about the Type of Timestamp Field
# the slave will use when transferring DAQ Packets to the master.
# The master has to use the same Type of Timestamp Field when transferring
# STIM Packets to the slave. TIMESTAMP_MODE and TIMEPSTAMP_TICKS contain
# information on the resolution of the data transfer clock.
def get_timestamp_length():
    return conf.contribs['XCP']['timestamp_size']


def identification_field_needs_alignment():
    try:
        identification_field_type_0 = conf.contribs['XCP'][
            'identification_field_type_0']
        identification_field_type_1 = conf.contribs['XCP'][
            'identification_field_type_1']
        if identification_field_type_1 == 1 and \
                identification_field_type_0 == 1:
            # relative odt with daq as word (aligned)
            return True
        return False
    except KeyError:
        return False


def get_daq_length():
    try:
        identification_field_type_0 = conf.contribs['XCP'][
            'identification_field_type_0']
        identification_field_type_1 = conf.contribs['XCP'][
            'identification_field_type_1']

        if identification_field_type_1 == 0 and \
                identification_field_type_0 == 0:
            # absolute odt number
            return 0
        if identification_field_type_1 == 0 and \
                identification_field_type_0 == 1:
            # relative odt with daq as byte
            return 1
        # relative odt with daq as word
        return 2
    except KeyError:
        return 0


def get_daq_data_field_length():
    try:
        data_length = get_max_dto()
    except KeyError:
        return 0
    data_length -= 1  # pid
    if identification_field_needs_alignment():
        data_length -= 1
    data_length -= get_daq_length()

    return data_length


# Idea taken from scapy/scapy/contrib/dce_rpc.py
class XCPEndiannessField(object):
    """Field which changes the endianness of a sub-field"""
    __slots__ = ["fld"]

    def __init__(self, fld):
        self.fld = fld

    def set_endianness(self):
        """Add the endianness to the format"""
        byte_oder = conf.contribs['XCP']['byte_order']
        endianness = ">" if byte_oder == 1 else "<"

        self.fld.fmt = endianness + self.fld.fmt[1:]
        self.fld.struct = struct.Struct(self.fld.fmt)

    def getfield(self, pkt, s):
        self.set_endianness()

        return self.fld.getfield(pkt, s)

    def addfield(self, pkt, s, val):
        self.set_endianness()
        return self.fld.addfield(pkt, s, val)

    def __getattr__(self, attr):
        return getattr(self.fld, attr)


class StrVarLenField(StrLenField):
    def randval(self):
        return RandBin(RandNum(0, self.max_length() or 1200))



# SPDX-License-Identifier: GPL-2.0-only
# This file is part of Scapy RPC
# See https://scapy.net/ for more information
# Copyright (C) Gabriel Potter

# ms-wkst.idl compiled on 06/07/2025
# This file is a stripped version ! Use scapy-rpc for the full.
"""
RPC definitions for the following interfaces:
- wkssvc (v1.0): 6BFFD098-A112-3610-9833-46C3F87E345A
This file is auto-generated by midl-to-scapy, do not modify.
"""

from enum import IntEnum
import uuid

from scapy.fields import StrFixedLenField
from scapy.layers.dcerpc import (
    NDRPacket,
    DceRpcOp,
    NDRConfPacketListField,
    NDRConfVarStrLenField,
    NDRConfVarStrLenFieldUtf16,
    NDRConfVarStrNullField,
    NDRConfVarStrNullFieldUtf16,
    NDRFullEmbPointerField,
    NDRFullPointerField,
    NDRInt3264EnumField,
    NDRIntField,
    NDRPacketField,
    NDRShortField,
    NDRSignedIntField,
    NDRUnionField,
    register_dcerpc_interface,
)


class LPWKSTA_INFO_100(NDRPacket):
    ALIGNMENT = (4, 8)
    fields_desc = [
        NDRIntField("wki100_platform_id", 0),
        NDRFullEmbPointerField(NDRConfVarStrNullFieldUtf16("wki100_computername", "")),
        NDRFullEmbPointerField(NDRConfVarStrNullFieldUtf16("wki100_langroup", "")),
        NDRIntField("wki100_ver_major", 0),
        NDRIntField("wki100_ver_minor", 0),
    ]


class LPWKSTA_INFO_101(NDRPacket):
    ALIGNMENT = (4, 8)
    fields_desc = [
        NDRIntField("wki101_platform_id", 0),
        NDRFullEmbPointerField(NDRConfVarStrNullFieldUtf16("wki101_computername", "")),
        NDRFullEmbPointerField(NDRConfVarStrNullFieldUtf16("wki101_langroup", "")),
        NDRIntField("wki101_ver_major", 0),
        NDRIntField("wki101_ver_minor", 0),
        NDRFullEmbPointerField(NDRConfVarStrNullFieldUtf16("wki101_lanroot", "")),
    ]


class LPWKSTA_INFO_102(NDRPacket):
    ALIGNMENT = (4, 8)
    fields_desc = [
        NDRIntField("wki102_platform_id", 0),
        NDRFullEmbPointerField(NDRConfVarStrNullFieldUtf16("wki102_computername", "")),
        NDRFullEmbPointerField(NDRConfVarStrNullFieldUtf16("wki102_langroup", "")),
        NDRIntField("wki102_ver_major", 0),
        NDRIntField("wki102_ver_minor", 0),
        NDRFullEmbPointerField(NDRConfVarStrNullFieldUtf16("wki102_lanroot", "")),
        NDRIntField("wki102_logged_on_users", 0),
    ]


class LPWKSTA_INFO_502(NDRPacket):
    ALIGNMENT = (4, 4)
    fields_desc = [
        NDRIntField("wki502_char_wait", 0),
        NDRIntField("wki502_collection_time", 0),
        NDRIntField("wki502_maximum_collection_count", 0),
        NDRIntField("wki502_keep_conn", 0),
        NDRIntField("wki502_max_cmds", 0),
        NDRIntField("wki502_sess_timeout", 0),
        NDRIntField("wki502_siz_char_buf", 0),
        NDRIntField("wki502_max_threads", 0),
        NDRIntField("wki502_lock_quota", 0),
        NDRIntField("wki502_lock_increment", 0),
        NDRIntField("wki502_lock_maximum", 0),
        NDRIntField("wki502_pipe_increment", 0),
        NDRIntField("wki502_pipe_maximum", 0),
        NDRIntField("wki502_cache_file_timeout", 0),
        NDRIntField("wki502_dormant_file_limit", 0),
        NDRIntField("wki502_read_ahead_throughput", 0),
        NDRIntField("wki502_num_mailslot_buffers", 0),
        NDRIntField("wki502_num_srv_announce_buffers", 0),
        NDRIntField("wki502_max_illegal_datagram_events", 0),
        NDRIntField("wki502_illegal_datagram_event_reset_frequency", 0),
        NDRSignedIntField("wki502_log_election_packets", 0),
        NDRSignedIntField("wki502_use_opportunistic_locking", 0),
        NDRSignedIntField("wki502_use_unlock_behind", 0),
        NDRSignedIntField("wki502_use_close_behind", 0),
        NDRSignedIntField("wki502_buf_named_pipes", 0),
        NDRSignedIntField("wki502_use_lock_read_unlock", 0),
        NDRSignedIntField("wki502_utilize_nt_caching", 0),
        NDRSignedIntField("wki502_use_raw_read", 0),
        NDRSignedIntField("wki502_use_raw_write", 0),
        NDRSignedIntField("wki502_use_write_raw_data", 0),
        NDRSignedIntField("wki502_use_encryption", 0),
        NDRSignedIntField("wki502_buf_files_deny_write", 0),
        NDRSignedIntField("wki502_buf_read_only_files", 0),
        NDRSignedIntField("wki502_force_core_create_mode", 0),
        NDRSignedIntField("wki502_use_512_byte_max_transfer", 0),
    ]


class LPWKSTA_INFO_1013(NDRPacket):
    ALIGNMENT = (4, 4)
    fields_desc = [NDRIntField("wki1013_keep_conn", 0)]


class LPWKSTA_INFO_1018(NDRPacket):
    ALIGNMENT = (4, 4)
    fields_desc = [NDRIntField("wki1018_sess_timeout", 0)]


class LPWKSTA_INFO_1046(NDRPacket):
    ALIGNMENT = (4, 4)
    fields_desc = [NDRIntField("wki1046_dormant_file_limit", 0)]


class NetrWkstaGetInfo_Request(NDRPacket):
    fields_desc = [
        NDRFullPointerField(NDRConfVarStrNullFieldUtf16("ServerName", "")),
        NDRIntField("Level", 0),
    ]


class NetrWkstaGetInfo_Response(NDRPacket):
    fields_desc = [
        NDRUnionField(
            [
                (
                    NDRFullPointerField(
                        NDRPacketField(
                            "WkstaInfo", LPWKSTA_INFO_100(), LPWKSTA_INFO_100
                        )
                    ),
                    (
                        (lambda pkt: getattr(pkt, "Level", None) == 100),
                        (lambda _, val: val.tag == 100),
                    ),
                ),
                (
                    NDRFullPointerField(
                        NDRPacketField(
                            "WkstaInfo", LPWKSTA_INFO_101(), LPWKSTA_INFO_101
                        )
                    ),
                    (
                        (lambda pkt: getattr(pkt, "Level", None) == 101),
                        (lambda _, val: val.tag == 101),
                    ),
                ),
                (
                    NDRFullPointerField(
                        NDRPacketField(
                            "WkstaInfo", LPWKSTA_INFO_102(), LPWKSTA_INFO_102
                        )
                    ),
                    (
                        (lambda pkt: getattr(pkt, "Level", None) == 102),
                        (lambda _, val: val.tag == 102),
                    ),
                ),
                (
                    NDRFullPointerField(
                        NDRPacketField(
                            "WkstaInfo", LPWKSTA_INFO_502(), LPWKSTA_INFO_502
                        )
                    ),
                    (
                        (lambda pkt: getattr(pkt, "Level", None) == 502),
                        (lambda _, val: val.tag == 502),
                    ),
                ),
                (
                    NDRFullPointerField(
                        NDRPacketField(
                            "WkstaInfo", LPWKSTA_INFO_1013(), LPWKSTA_INFO_1013
                        )
                    ),
                    (
                        (lambda pkt: getattr(pkt, "Level", None) == 1013),
                        (lambda _, val: val.tag == 1013),
                    ),
                ),
                (
                    NDRFullPointerField(
                        NDRPacketField(
                            "WkstaInfo", LPWKSTA_INFO_1018(), LPWKSTA_INFO_1018
                        )
                    ),
                    (
                        (lambda pkt: getattr(pkt, "Level", None) == 1018),
                        (lambda _, val: val.tag == 1018),
                    ),
                ),
                (
                    NDRFullPointerField(
                        NDRPacketField(
                            "WkstaInfo", LPWKSTA_INFO_1046(), LPWKSTA_INFO_1046
                        )
                    ),
                    (
                        (lambda pkt: getattr(pkt, "Level", None) == 1046),
                        (lambda _, val: val.tag == 1046),
                    ),
                ),
            ],
            StrFixedLenField("WkstaInfo", "", length=0),
            align=(4, 8),
            switch_fmt=("L", "L"),
        ),
        NDRIntField("status", 0),
    ]


class NET_COMPUTER_NAME_TYPE(IntEnum):
    NetPrimaryComputerName = 0
    NetAlternateComputerNames = 1
    NetAllComputerNames = 2
    NetComputerNameTypeMax = 3


class PUNICODE_STRING(NDRPacket):
    ALIGNMENT = (4, 8)
    fields_desc = [
        NDRShortField("Length", None, size_of="Buffer", adjust=lambda _, x: (x * 2)),
        NDRShortField(
            "MaximumLength", None, size_of="Buffer", adjust=lambda _, x: (x * 2)
        ),
        NDRFullEmbPointerField(
            NDRConfVarStrLenFieldUtf16(
                "Buffer",
                "",
                size_is=lambda pkt: (pkt.MaximumLength // 2),
                length_is=lambda pkt: (pkt.Length // 2),
            )
        ),
    ]


class PNET_COMPUTER_NAME_ARRAY(NDRPacket):
    ALIGNMENT = (4, 8)
    fields_desc = [
        NDRIntField("EntryCount", None, size_of="ComputerNames"),
        NDRFullEmbPointerField(
            NDRConfPacketListField(
                "ComputerNames",
                [PUNICODE_STRING()],
                PUNICODE_STRING,
                size_is=lambda pkt: pkt.EntryCount,
            )
        ),
    ]


class NetrEnumerateComputerNames_Request(NDRPacket):
    fields_desc = [
        NDRFullPointerField(NDRConfVarStrNullFieldUtf16("ServerName", "")),
        NDRInt3264EnumField("NameType", 0, NET_COMPUTER_NAME_TYPE),
        NDRIntField("Reserved", 0),
    ]


class NetrEnumerateComputerNames_Response(NDRPacket):
    fields_desc = [
        NDRFullPointerField(
            NDRPacketField(
                "ComputerNames", PNET_COMPUTER_NAME_ARRAY(), PNET_COMPUTER_NAME_ARRAY
            )
        ),
        NDRIntField("status", 0),
    ]


WKSSVC_OPNUMS = {
    0: DceRpcOp(NetrWkstaGetInfo_Request, NetrWkstaGetInfo_Response),
    # 3: Opnum3NotUsedOnWire,
    # 4: Opnum4NotUsedOnWire,
    # 12: Opnum12NotUsedOnWire,
    # 14: Opnum14NotUsedOnWire,
    # 15: Opnum15NotUsedOnWire,
    # 16: Opnum16NotUsedOnWire,
    # 17: Opnum17NotUsedOnWire,
    # 18: Opnum18NotUsedOnWire,
    # 19: Opnum19NotUsedOnWire,
    # 21: Opnum21NotUsedOnWire,
    30: DceRpcOp(
        NetrEnumerateComputerNames_Request, NetrEnumerateComputerNames_Response
    ),
}
register_dcerpc_interface(
    name="wkssvc",
    uuid=uuid.UUID("6BFFD098-A112-3610-9833-46C3F87E345A"),
    version="1.0",
    opnums=WKSSVC_OPNUMS,
)



# SPDX-License-Identifier: GPL-2.0-only
# This file is part of Scapy
# See https://scapy.net/ for more information
# Copyright (C) Lucas Preston <lucas.preston@infinite.io>

# scapy.contrib.description = Portmapper v2
# scapy.contrib.status = loads

from scapy.packet import Packet, bind_layers
from scapy.fields import IntField, PacketListField
from scapy.contrib.oncrpc import RPC, RPC_Call


class GETPORT_Call(Packet):
    name = 'GETPORT Call'
    fields_desc = [
        IntField('prog', 0),
        IntField('vers', 0),
        IntField('prot', 0),
        IntField('port', 0)
    ]


class GETPORT_Reply(Packet):
    name = 'GETPORT Reply'
    fields_desc = [
        IntField('port', 0)
    ]


bind_layers(RPC, GETPORT_Call, mtype=0)
bind_layers(RPC, GETPORT_Reply, mtype=1)
bind_layers(
    RPC_Call, GETPORT_Call, program=100000, pversion=2, procedure=3
)


class NULL_Call(Packet):
    name = 'PORTMAP NULL Call'
    fields_desc = []


class NULL_Reply(Packet):
    name = 'PORTMAP NULL Reply'
    fields_desc = []


bind_layers(RPC, NULL_Call, mtype=0)
bind_layers(RPC, NULL_Reply, mtype=1)
bind_layers(RPC_Call, NULL_Call, program=100000, pversion=2, procedure=0)


class Map_Entry(Packet):
    name = 'PORTMAP Map Entry'
    fields_desc = [
        IntField('prog', 0),
        IntField('vers', 0),
        IntField('prot', 0),
        IntField('port', 0),
        IntField('value_follows', 0)
    ]

    def extract_padding(self, s):
        return '', s


class DUMP_Call(Packet):
    name = 'PORTMAP DUMP Call'
    fields_desc = []


class DUMP_Reply(Packet):
    name = 'PORTMAP DUMP Reply'
    fields_desc = [
        IntField('value_follows', 0),
        PacketListField('mappings', [], Map_Entry,
                        next_cls_cb=lambda pkt, lst, cur, remain:
                        Map_Entry if pkt.value_follows == 1 and
                        (len(lst) == 0 or cur.value_follows == 1) and
                        len(remain) > 4 else None)
    ]


bind_layers(RPC, DUMP_Call, mtype=0)
bind_layers(RPC, DUMP_Reply, mtype=1)
bind_layers(RPC_Call, DUMP_Call, program=100000, pversion=2, procedure=4)



# SPDX-License-Identifier: GPL-2.0-or-later
# This file is part of Scapy
# See https://scapy.net/ for more information
# Copyright (C) 2009 Jochen Bartl

# scapy.contrib.description = Enhanced Interior Gateway Routing Protocol (EIGRP)
# scapy.contrib.status = loads

"""
    EIGRP Scapy Extension
    ~~~~~~~~~~~~~~~~~~~~~

    :version:   2009-08-13
    :e-mail:    lobo@c3a.de / jochen.bartl@gmail.com

    :TODO

    - Replace TLV code with a more generic solution
        * http://trac.secdev.org/scapy/ticket/90
    - Write function for calculating authentication data

    :Thanks:

    - TLV code derived from the CDP implementation of scapy. (Thanks to Nicolas Bareil and Arnaud Ebalard)
        http://trac.secdev.org/scapy/ticket/18
    - IOS / EIGRP Version Representation FIX by Dirk Loss
"""
import socket
import struct

from scapy.packet import Packet
from scapy.fields import StrField, IPField, XShortField, FieldLenField, \
    StrLenField, IntField, ByteEnumField, ByteField, ConditionalField, \
    FlagsField, IP6Field, PacketListField, ShortEnumField, \
    ShortField, StrFixedLenField, ThreeBytesField
from scapy.layers.inet import IP, checksum, bind_layers
from scapy.layers.inet6 import IPv6
from scapy.compat import chb
from scapy.config import conf
from scapy.utils import inet_aton, inet_ntoa
from scapy.pton_ntop import inet_ntop, inet_pton
from scapy.error import warning, Scapy_Exception
from scapy.volatile import RandShort, RandString


class EigrpIPField(StrField, IPField):
    """
    This is a special field type for handling ip addresses of destination networks in internal and
    external route updates.

    EIGRP removes zeros from the host portion of the ip address if the netmask is 8, 16 or 24 bits.
    """

    __slots__ = ["length_from"]

    def __init__(self, name, default, length=None, length_from=None):
        StrField.__init__(self, name, default)
        self.length_from = length_from
        if length is not None:
            self.length_from = lambda pkt, length=length: length

    def h2i(self, pkt, x):
        return IPField.h2i(self, pkt, x)

    def i2m(self, pkt, x):
        x = inet_aton(x)
        tmp_len = self.length_from(pkt)

        if tmp_len <= 8:
            return x[:1]
        elif tmp_len <= 16:
            return x[:2]
        elif tmp_len <= 24:
            return x[:3]
        else:
            return x

    def m2i(self, pkt, x):
        tmp_len = self.length_from(pkt)

        if tmp_len <= 8:
            x += b"\x00\x00\x00"
        elif tmp_len <= 16:
            x += b"\x00\x00"
        elif tmp_len <= 24:
            x += b"\x00"

        return inet_ntoa(x)

    def prefixlen_to_bytelen(self, tmp_len):
        if tmp_len <= 8:
            tmp_len = 1
        elif tmp_len <= 16:
            tmp_len = 2
        elif tmp_len <= 24:
            tmp_len = 3
        else:
            tmp_len = 4

        return tmp_len

    def i2len(self, pkt, x):
        tmp_len = self.length_from(pkt)
        tmp_len = self.prefixlen_to_bytelen(tmp_len)
        return tmp_len

    def getfield(self, pkt, s):
        tmp_len = self.length_from(pkt)
        tmp_len = self.prefixlen_to_bytelen(tmp_len)
        return s[tmp_len:], self.m2i(pkt, s[:tmp_len])

    def randval(self):
        return IPField.randval(self)


class EigrpIP6Field(StrField, IP6Field):
    """
    This is a special field type for handling ip addresses of destination networks in internal and
    external route updates.

    """

    __slots__ = ["length_from"]

    def __init__(self, name, default, length=None, length_from=None):
        StrField.__init__(self, name, default)
        self.length_from = length_from
        if length is not None:
            self.length_from = lambda pkt, length=length: length

    def any2i(self, pkt, x):
        return IP6Field.any2i(self, pkt, x)

    def i2repr(self, pkt, x):
        return IP6Field.i2repr(self, pkt, x)

    def h2i(self, pkt, x):
        return IP6Field.h2i(self, pkt, x)

    def i2m(self, pkt, x):
        x = inet_pton(socket.AF_INET6, x)
        tmp_len = self.length_from(pkt)
        tmp_len = self.prefixlen_to_bytelen(tmp_len)

        return x[:tmp_len]

    def m2i(self, pkt, x):
        tmp_len = self.length_from(pkt)

        prefixlen = self.prefixlen_to_bytelen(tmp_len)
        if tmp_len > 128:
            warning("EigrpIP6Field: Prefix length is > 128. Dissection of this packet will fail")  # noqa: E501
        else:
            pad = b"\x00" * (16 - prefixlen)
            x += pad

        return inet_ntop(socket.AF_INET6, x)

    def prefixlen_to_bytelen(self, plen):
        plen = plen // 8

        if plen < 16:
            plen += 1

        return plen

    def i2len(self, pkt, x):
        tmp_len = self.length_from(pkt)
        tmp_len = self.prefixlen_to_bytelen(tmp_len)
        return tmp_len

    def getfield(self, pkt, s):
        tmp_len = self.length_from(pkt)
        tmp_len = self.prefixlen_to_bytelen(tmp_len)
        return s[tmp_len:], self.m2i(pkt, s[:tmp_len])

    def randval(self):
        return IP6Field.randval(self)


class EIGRPGeneric(Packet):
    name = "EIGRP Generic TLV"
    fields_desc = [XShortField("type", 0x0000),
                   FieldLenField("len", None, "value", "!H", adjust=lambda pkt, x: x + 4),  # noqa: E501
                   StrLenField("value", b"\x00", length_from=lambda pkt: pkt.len - 4)]  # noqa: E501

    def guess_payload_class(self, p):
        return conf.padding_layer


class EIGRPParam(EIGRPGeneric):
    name = "EIGRP Parameters"
    fields_desc = [XShortField("type", 0x0001),
                   ShortField("len", 12),
                   # Bandwidth
                   ByteField("k1", 1),
                   # Load
                   ByteField("k2", 0),
                   # Delay
                   ByteField("k3", 1),
                   # Reliability
                   ByteField("k4", 0),
                   # MTU
                   ByteField("k5", 0),
                   ByteField("reserved", 0),
                   ShortField("holdtime", 15)
                   ]


class EIGRPAuthData(EIGRPGeneric):
    name = "EIGRP Authentication Data"
    fields_desc = [XShortField("type", 0x0002),
                   FieldLenField("len", None, "authdata", "!H", adjust=lambda pkt, x: x + 24),  # noqa: E501
                   ShortEnumField("authtype", 2, {2: "MD5"}),
                   ShortField("keysize", None),
                   IntField("keyid", 1),
                   StrFixedLenField("nullpad", b"\x00" * 12, 12),
                   StrLenField("authdata", RandString(16), length_from=lambda pkt: pkt.keysize)  # noqa: E501
                   ]

    def post_build(self, p, pay):
        p += pay

        if self.keysize is None:
            keysize = len(self.authdata)
            p = p[:6] + chb((keysize >> 8) & 0xff) + chb(keysize & 0xff) + p[8:]  # noqa: E501

        return p


class EIGRPSeq(EIGRPGeneric):
    name = "EIGRP Sequence"
    fields_desc = [XShortField("type", 0x0003),
                   ShortField("len", None),
                   ByteField("addrlen", 4),
                   ConditionalField(IPField("ipaddr", "192.168.0.1"),
                                    lambda pkt:pkt.addrlen == 4),
                   ConditionalField(IP6Field("ip6addr", "2001::"),
                                    lambda pkt:pkt.addrlen == 16)
                   ]

    def post_build(self, p, pay):
        p += pay

        if self.len is None:
            tmp_len = len(p)
            tmp_p = p[:2] + chb((tmp_len >> 8) & 0xff)
            p = tmp_p + chb(tmp_len & 0xff) + p[4:]

        return p


class ShortVersionField(ShortField):
    def i2repr(self, pkt, x):
        try:
            minor = x & 0xff
            major = (x >> 8) & 0xff
        except TypeError:
            return "unknown"
        else:
            # We print a leading 'v' so that these values don't look like floats  # noqa: E501
            return "v%s.%s" % (major, minor)

    def h2i(self, pkt, x):
        """The field accepts string values like v12.1, v1.1 or integer values.
           String values have to start with a "v" followed by a
           floating point number. Valid numbers are between 0 and 255.

        """

        if isinstance(x, str) and x.startswith("v") and len(x) <= 8:
            major = int(x.split(".")[0][1:])
            minor = int(x.split(".")[1])

            return (major << 8) | minor

        elif isinstance(x, int) and 0 <= x <= 65535:
            return x
        else:
            if not hasattr(self, "default"):
                return x
            if self.default is not None:
                warning("set value to default. Format of %r is invalid", x)
                return self.default
            else:
                raise Scapy_Exception("Format of value is invalid")

    def randval(self):
        return RandShort()


class EIGRPSwVer(EIGRPGeneric):
    name = "EIGRP Software Version"
    fields_desc = [XShortField("type", 0x0004),
                   ShortField("len", 8),
                   ShortVersionField("ios", "v12.0"),
                   ShortVersionField("eigrp", "v1.2")
                   ]


class EIGRPNms(EIGRPGeneric):
    name = "EIGRP Next Multicast Sequence"
    fields_desc = [XShortField("type", 0x0005),
                   ShortField("len", 8),
                   IntField("nms", 2)
                   ]


# Don't get confused by the term "receive-only". This flag is always set, when you configure  # noqa: E501
# one of the stub options. It's also the only flag set, when you configure "eigrp stub receive-only".  # noqa: E501
_EIGRP_STUB_FLAGS = ["connected", "static", "summary", "receive-only", "redistributed", "leak-map"]  # noqa: E501


class EIGRPStub(EIGRPGeneric):
    name = "EIGRP Stub Router"
    fields_desc = [XShortField("type", 0x0006),
                   ShortField("len", 6),
                   FlagsField("flags", 0x000d, 16, _EIGRP_STUB_FLAGS)]

# Delay 0xffffffff == Destination Unreachable


class EIGRPIntRoute(EIGRPGeneric):
    name = "EIGRP Internal Route"
    fields_desc = [XShortField("type", 0x0102),
                   FieldLenField("len", None, "dst", "!H", adjust=lambda pkt, x: x + 25),  # noqa: E501
                   IPField("nexthop", "192.168.0.0"),
                   IntField("delay", 128000),
                   IntField("bandwidth", 256),
                   ThreeBytesField("mtu", 1500),
                   ByteField("hopcount", 0),
                   ByteField("reliability", 255),
                   ByteField("load", 0),
                   XShortField("reserved", 0),
                   ByteField("prefixlen", 24),
                   EigrpIPField("dst", "192.168.1.0", length_from=lambda pkt: pkt.prefixlen),  # noqa: E501
                   ]


_EIGRP_EXTERNAL_PROTOCOL_ID = {
    0x01: "IGRP",
    0x02: "EIGRP",
    0x03: "Static Route",
    0x04: "RIP",
    0x05: "Hello",
    0x06: "OSPF",
    0x07: "IS-IS",
    0x08: "EGP",
    0x09: "BGP",
    0x0A: "IDRP",
    0x0B: "Connected Link"
}

_EIGRP_EXTROUTE_FLAGS = ["external", "candidate-default"]


class EIGRPExtRoute(EIGRPGeneric):
    name = "EIGRP External Route"
    fields_desc = [XShortField("type", 0x0103),
                   FieldLenField("len", None, "dst", "!H", adjust=lambda pkt, x: x + 45),  # noqa: E501
                   IPField("nexthop", "192.168.0.0"),
                   IPField("originrouter", "192.168.0.1"),
                   IntField("originasn", 0),
                   IntField("tag", 0),
                   IntField("externalmetric", 0),
                   ShortField("reserved", 0),
                   ByteEnumField("extprotocolid", 3, _EIGRP_EXTERNAL_PROTOCOL_ID),  # noqa: E501
                   FlagsField("flags", 0, 8, _EIGRP_EXTROUTE_FLAGS),
                   IntField("delay", 0),
                   IntField("bandwidth", 256),
                   ThreeBytesField("mtu", 1500),
                   ByteField("hopcount", 0),
                   ByteField("reliability", 255),
                   ByteField("load", 0),
                   XShortField("reserved2", 0),
                   ByteField("prefixlen", 24),
                   EigrpIPField("dst", "192.168.1.0", length_from=lambda pkt: pkt.prefixlen)  # noqa: E501
                   ]


class EIGRPv6IntRoute(EIGRPGeneric):
    name = "EIGRP for IPv6 Internal Route"
    fields_desc = [XShortField("type", 0x0402),
                   FieldLenField("len", None, "dst", "!H", adjust=lambda pkt, x: x + 37),  # noqa: E501
                   IP6Field("nexthop", "::"),
                   IntField("delay", 128000),
                   IntField("bandwidth", 256000),
                   ThreeBytesField("mtu", 1500),
                   ByteField("hopcount", 1),
                   ByteField("reliability", 255),
                   ByteField("load", 0),
                   XShortField("reserved", 0),
                   ByteField("prefixlen", 16),
                   EigrpIP6Field("dst", "2001::", length_from=lambda pkt: pkt.prefixlen)  # noqa: E501
                   ]


class EIGRPv6ExtRoute(EIGRPGeneric):
    name = "EIGRP for IPv6 External Route"
    fields_desc = [XShortField("type", 0x0403),
                   FieldLenField("len", None, "dst", "!H", adjust=lambda pkt, x: x + 57),  # noqa: E501
                   IP6Field("nexthop", "::"),
                   IPField("originrouter", "192.168.0.1"),
                   IntField("originasn", 0),
                   IntField("tag", 0),
                   IntField("externalmetric", 0),
                   ShortField("reserved", 0),
                   ByteEnumField("extprotocolid", 3, _EIGRP_EXTERNAL_PROTOCOL_ID),  # noqa: E501
                   FlagsField("flags", 0, 8, _EIGRP_EXTROUTE_FLAGS),
                   IntField("delay", 0),
                   IntField("bandwidth", 256000),
                   ThreeBytesField("mtu", 1500),
                   ByteField("hopcount", 1),
                   ByteField("reliability", 0),
                   ByteField("load", 1),
                   XShortField("reserved2", 0),
                   ByteField("prefixlen", 8),
                   EigrpIP6Field("dst", "::", length_from=lambda pkt: pkt.prefixlen)  # noqa: E501
                   ]


_eigrp_tlv_cls = {
    0x0001: "EIGRPParam",
    0x0002: "EIGRPAuthData",
    0x0003: "EIGRPSeq",
    0x0004: "EIGRPSwVer",
    0x0005: "EIGRPNms",
    0x0006: "EIGRPStub",
    0x0102: "EIGRPIntRoute",
    0x0103: "EIGRPExtRoute",
    0x0402: "EIGRPv6IntRoute",
    0x0403: "EIGRPv6ExtRoute"
}


def _EIGRPGuessPayloadClass(p, **kargs):
    cls = conf.raw_layer
    if len(p) >= 2:
        t = struct.unpack("!H", p[:2])[0]
        clsname = _eigrp_tlv_cls.get(t, "EIGRPGeneric")
        cls = globals()[clsname]
    return cls(p, **kargs)


_EIGRP_OPCODES = {1: "Update",
                  2: "Request",
                  3: "Query",
                  4: "Replay",
                  5: "Hello",
                  6: "IPX SAP",
                  10: "SIA Query",
                  11: "SIA Reply"}

# The Conditional Receive bit is used for reliable multicast communication.
# Update-Flag: Not sure if Cisco calls it that way, but it's set when neighbors
# are exchanging routing information
_EIGRP_FLAGS = ["init", "cond-recv", "unknown", "update"]


class EIGRP(Packet):
    name = "EIGRP"
    fields_desc = [ByteField("ver", 2),
                   ByteEnumField("opcode", 5, _EIGRP_OPCODES),
                   XShortField("chksum", None),
                   FlagsField("flags", 0, 32, _EIGRP_FLAGS),
                   IntField("seq", 0),
                   IntField("ack", 0),
                   IntField("asn", 100),
                   PacketListField("tlvlist", [], _EIGRPGuessPayloadClass)
                   ]

    def post_build(self, p, pay):
        p += pay
        if self.chksum is None:
            c = checksum(p)
            p = p[:2] + chb((c >> 8) & 0xff) + chb(c & 0xff) + p[4:]
        return p

    def mysummary(self):
        summarystr = "EIGRP (AS=%EIGRP.asn% Opcode=%EIGRP.opcode%"
        if self.opcode == 5 and self.ack != 0:
            summarystr += " (ACK)"
        if self.flags != 0:
            summarystr += " Flags=%EIGRP.flags%"

        return self.sprintf(summarystr + ")")


bind_layers(IP, EIGRP, proto=88)
bind_layers(IPv6, EIGRP, nh=88)



# SPDX-License-Identifier: GPL-2.0-only
# This file is part of Scapy
# See https://scapy.net/ for more information
# Copyright (C) Andreas Korb <andreas.d.korb@gmail.com>
# Copyright (C) Nils Weiss <nils@we155.de>

# scapy.contrib.status = skip

from scapy.fields import ByteField, XByteField, BitEnumField, \
    PacketListField, XBitField, XByteEnumField, FieldListField, FieldLenField
from scapy.packet import Packet
from scapy.contrib.automotive.obd.packet import OBD_Packet
from scapy.config import conf


class OBD_DTC(OBD_Packet):
    name = "DiagnosticTroubleCode"

    locations = {
        0b00: 'Powertrain',
        0b01: 'Chassis',
        0b10: 'Body',
        0b11: 'Network',
    }

    fields_desc = [
        BitEnumField('location', 0, 2, locations),
        XBitField('code1', 0, 2),
        XBitField('code2', 0, 4),
        XBitField('code3', 0, 4),
        XBitField('code4', 0, 4),
    ]


class OBD_NR(Packet):
    name = "NegativeResponse"

    responses = {
        0x10: 'generalReject',
        0x11: 'serviceNotSupported',
        0x12: 'subFunctionNotSupported-InvalidFormat',
        0x21: 'busy-RepeatRequest',
        0x22: 'conditionsNotCorrectOrRequestSequenceError',
        0x78: 'requestCorrectlyReceived-ResponsePending'
    }

    fields_desc = [
        XByteField('request_service_id', 0),
        XByteEnumField('response_code', 0, responses)
    ]

    def answers(self, other):
        return self.request_service_id == other.service and \
            (self.response_code != 0x78 or
             conf.contribs['OBD']['treat-response-pending-as-answer'])


class OBD_S01(Packet):
    name = "S1_CurrentData"
    fields_desc = [
        FieldListField("pid", [], XByteField('', 0))
    ]


class OBD_S02_Record(OBD_Packet):
    fields_desc = [
        XByteField('pid', 0),
        ByteField('frame_no', 0)
    ]


class OBD_S02(Packet):
    name = "S2_FreezeFrameData"
    fields_desc = [
        PacketListField("requests", [], OBD_S02_Record)
    ]


class OBD_S03(Packet):
    name = "S3_RequestDTCs"


class OBD_S03_PR(Packet):
    name = "S3_ResponseDTCs"
    fields_desc = [
        FieldLenField('count', None, count_of='dtcs', fmt='B'),
        PacketListField('dtcs', [], OBD_DTC, count_from=lambda pkt: pkt.count)
    ]

    def answers(self, other):
        return isinstance(other, OBD_S03)


class OBD_S04(Packet):
    name = "S4_ClearDTCs"


class OBD_S04_PR(Packet):
    name = "S4_ClearDTCsPositiveResponse"

    def answers(self, other):
        return isinstance(other, OBD_S04)


class OBD_S06(Packet):
    name = "S6_OnBoardDiagnosticMonitoring"
    fields_desc = [
        FieldListField("mid", [], XByteField('', 0))
    ]


class OBD_S07(Packet):
    name = "S7_RequestPendingDTCs"


class OBD_S07_PR(Packet):
    name = "S7_ResponsePendingDTCs"
    fields_desc = [
        FieldLenField('count', None, count_of='dtcs', fmt='B'),
        PacketListField('dtcs', [], OBD_DTC, count_from=lambda pkt: pkt.count)
    ]

    def answers(self, other):
        return isinstance(other, OBD_S07)


class OBD_S08(Packet):
    name = "S8_RequestControlOfSystem"
    fields_desc = [
        FieldListField("tid", [], XByteField('', 0))
    ]


class OBD_S09(Packet):
    name = "S9_VehicleInformation"
    fields_desc = [
        FieldListField("iid", [], XByteField('', 0))
    ]


class OBD_S0A(Packet):
    name = "S0A_RequestPermanentDTCs"


class OBD_S0A_PR(Packet):
    name = "S0A_ResponsePermanentDTCs"
    fields_desc = [
        FieldLenField('count', None, count_of='dtcs', fmt='B'),
        PacketListField('dtcs', [], OBD_DTC, count_from=lambda pkt: pkt.count)
    ]

    def answers(self, other):
        return isinstance(other, OBD_S0A)
